################################################################################
# © Copyright 2021-2022 Zapata Computing Inc.
################################################################################
import numpy as np
import pytest
import sympy

from orquestra.quantum.circuits import (
    Circuit,
    CustomGateDefinition,
    GateOperation,
    MultiPhaseOperation,
    split_circuit,
)
from orquestra.quantum.circuits._builtin_gates import (
    CNOT,
    CPHASE,
    CZ,
    ISWAP,
    PHASE,
    RX,
    RY,
    RZ,
    SWAP,
    XX,
    YY,
    ZZ,
    H,
    I,
    T,
    X,
    Y,
    Z,
)

RNG = np.random.default_rng(42)

EXAMPLE_OPERATIONS = tuple(
    [
        *[gate(qubit_i) for qubit_i in [0, 1, 4] for gate in [X, Y, Z, H, I, T]],
        *[
            gate(angle)(qubit_i)
            for qubit_i in [0, 1, 4]
            for gate in [PHASE, RX, RY, RZ]
            for angle in [0, 0.1, np.pi / 5, np.pi, 2 * np.pi, sympy.Symbol("theta")]
        ],
        *[
            gate(qubit1_i, qubit2_i)
            for qubit1_i, qubit2_i in [(0, 1), (1, 0), (0, 5), (4, 2)]
            for gate in [CNOT, CZ, SWAP, ISWAP]
        ],
        *[
            gate(angle)(qubit1_i, qubit2_i)
            for qubit1_i, qubit2_i in [(0, 1), (1, 0), (0, 5), (4, 2)]
            for gate in [CPHASE, XX, YY, ZZ]
            for angle in [0, 0.1, np.pi / 5, np.pi, 2 * np.pi, sympy.Symbol("theta")]
        ],
    ]
)


class TestInitialization:
    def test_creating_circuit_has_correct_operations(self):
        circuit = Circuit(operations=EXAMPLE_OPERATIONS)
        assert circuit.operations == list(EXAMPLE_OPERATIONS)

    @pytest.mark.parametrize("n_qubits", [-1.3523, -2])
    def test_creating_circuit_with_negative_n_qubits_fails(self, n_qubits):
        with pytest.raises(ValueError):
            Circuit(n_qubits=n_qubits)

    @pytest.mark.parametrize("n_qubits", [1.3523, 2.292])
    def test_creating_circuit_with_float_n_qubits_fails(self, n_qubits):
        with pytest.raises(ValueError):
            circuit = Circuit(n_qubits=n_qubits)

            assert circuit.n_qubits == int(n_qubits)

    def test_creating_circuit_with_float_integer_passes(self):
        Circuit(n_qubits=5.0)


@pytest.mark.parametrize(
    "circuit",
    [
        Circuit(),
        Circuit([]),
        Circuit([H(0)]),
        Circuit([H(0)], 5),
    ],
)
def test_printing_circuit_doesnt_raise_exception(circuit):
    str(circuit)
    repr(circuit)


class TestConcatenation:
    def test_appending_to_circuit_yields_correct_operations(self):
        circuit = Circuit()
        circuit += H(0)
        circuit += CNOT(0, 2)

        assert circuit.operations == [H(0), CNOT(0, 2)]
        assert circuit.n_qubits == 3

    def test_circuits_sum_yields_correct_operations(self):
        circuit1 = Circuit()
        circuit1 += H(0)
        circuit1 += CNOT(0, 2)

        circuit2 = Circuit([X(2), YY(sympy.Symbol("theta"))(5)])

        res_circuit = circuit1 + circuit2
        assert res_circuit.operations == [
            H(0),
            CNOT(0, 2),
            X(2),
            YY(sympy.Symbol("theta"))(5),
        ]
        assert res_circuit.n_qubits == 6


class TestBindingParams:
    def test_circuit_bound_with_all_params_contains_bound_gates(self):
        theta1, theta2, theta3 = sympy.symbols("theta1:4")
        symbols_map = {theta1: 0.5, theta2: 3.14, theta3: 0}

        circuit = Circuit(
            [
                RX(theta1)(0),
                RY(theta2)(1),
                RZ(theta3)(0),
                RX(theta3)(0),
            ]
        )
        bound_circuit = circuit.bind(symbols_map)

        expected_circuit = Circuit(
            [
                RX(theta1).bind(symbols_map)(0),
                RY(theta2).bind(symbols_map)(1),
                RZ(theta3).bind(symbols_map)(0),
                RX(theta3).bind(symbols_map)(0),
            ]
        )

        assert bound_circuit == expected_circuit

    def test_binding_all_params_leaves_no_free_symbols(self):
        alpha, beta, gamma = sympy.symbols("alpha,beta,gamma")
        circuit = Circuit(
            [
                RX(alpha)(0),
                RY(beta)(1),
                RZ(gamma)(0),
                RX(gamma)(0),
            ]
        )
        bound_circuit = circuit.bind({alpha: 0.5, beta: 3.14, gamma: 0})

        assert not bound_circuit.free_symbols

    def test_binding_some_params_leaves_free_params(self):
        theta1, theta2, theta3 = sympy.symbols("theta1:4")
        circuit = Circuit(
            [
                RX(theta1)(0),
                RY(theta2)(1),
                RZ(theta3)(0),
                RX(theta2)(0),
            ]
        )

        bound_circuit = circuit.bind({theta1: 0.5, theta3: 3.14})
        assert bound_circuit.free_symbols == [theta2]

    def test_binding_excessive_params_binds_only_the_existing_ones(self):
        theta1, theta2, theta3 = sympy.symbols("theta1:4")
        other_param = sympy.symbols("other_param")
        circuit = Circuit(
            [
                RX(theta1)(0),
                RY(theta2)(1),
                RZ(theta3)(0),
                RX(theta2)(0),
            ]
        )

        bound_circuit = circuit.bind({theta1: -np.pi, other_param: 42})
        assert bound_circuit.free_symbols == [theta2, theta3]

    def test_symbols_of_wavefunction_operations_are_present_in_circuits_free_symbols(
        self,
    ):
        alpha, beta = sympy.symbols("alpha, beta")
        circuit = Circuit([RX(alpha)(0), MultiPhaseOperation((beta, 0.5))])

        assert circuit.free_symbols == [alpha, beta]

    def test_binding_symbols_to_circuit_binds_them_to_wavefunction_operation(self):
        alpha, beta = sympy.symbols("alpha, beta")
        circuit = Circuit(
            [RX(alpha)(0), MultiPhaseOperation((beta, 0.5)), RX(beta)(1)]
        ).bind({beta: 0.3})

        assert circuit.free_symbols == [alpha]
        assert circuit.operations == [
            RX(alpha)(0),
            MultiPhaseOperation((0.3, 0.5)),
            RX(0.3)(1),
        ]


def test_splitting_circuits_partitions_it_into_expected_chunks():
    def _predicate(operation):
        return isinstance(operation, GateOperation) and operation.gate.name in (
            "RX",
            "RY",
            "RZ",
        )

    circuit = Circuit(
        [RX(np.pi)(0), RZ(np.pi / 2)(1), CNOT(2, 3), RY(np.pi / 4)(2), X(0), Y(1)]
    )

    expected_partition = [
        (True, Circuit([RX(np.pi)(0), RZ(np.pi / 2)(1)], n_qubits=4)),
        (False, Circuit([CNOT(2, 3)], n_qubits=4)),
        (True, Circuit([RY(np.pi / 4)(2)], n_qubits=4)),
        (False, Circuit([X(0), Y(1)], n_qubits=4)),
    ]

    assert list(split_circuit(circuit, _predicate)) == expected_partition


class TestCircuitInverse:
    @pytest.mark.parametrize(
        "circuit",
        [
            Circuit([H(0)]),
            Circuit([H(1), CZ(0, 1), H(0)]),
            Circuit([RX(0.1)(0), RZ(sympy.pi)(0), RZ(0.1)(1)]),
        ],
    )
    def test_concatenating_circuit_and_its_ivnerse_yields_identity_circuit(
        self, circuit
    ):
        assert np.allclose(
            np.array((circuit + circuit.inverse()).to_unitary()),
            np.eye(2**circuit.n_qubits),
        )

    def test_inverting_an_empty_circuit_yields_an_empty_circuit(self):
        circuit = Circuit()
        circuit_inverse = circuit.inverse()
        assert circuit_inverse == circuit

    def test_inverting_a_circuit_without_dagger_fails(self):
        custom_a = CustomGateDefinition(
            gate_name="custom_a",
            matrix=sympy.Matrix(
                [
                    [-1, 0],
                    [0, 1],
                ]
            ),
            params_ordering=(),
        )
        gate = GateOperation(custom_a, (0,))
        circuit = Circuit(operations=[gate])
        with pytest.raises(AttributeError):
            circuit.inverse()



################################################################################
# © Copyright 2022 Zapata Computing Inc.
################################################################################
from .expectation_values import (
    ExpectationValues,
    concatenate_expectation_values,
    expectation_values_to_real,
    get_expectation_values_from_parities,
    load_expectation_values,
    save_expectation_values,
)
from .measurements import (
    Measurements,
    _check_sample_elimination,
    _convert_bitstrings_to_vector,
    convert_bitstring_to_int,
    get_expectation_value_from_frequencies,
)
from .parities import (
    Parities,
    check_parity,
    check_parity_of_vector,
    get_parities_from_measurements,
    load_parities,
    save_parities,
)



#   Copyright 2017 The OpenFermion Developers
#   Modifications copyright 2022 Zapata Computing, Inc. for compatibility reasons.
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""Tests for sparse_tools.py."""
import unittest

import numpy
import scipy
from scipy.sparse import csc_matrix

from orquestra.quantum.operators import (
    PauliSum,
    PauliTerm,
    expectation,
    get_sparse_operator,
)


class GetSparseOperatorTest(unittest.TestCase):
    def test_get_sparse_operator_n_qubits_too_small(self):
        with self.assertRaises(ValueError):
            get_sparse_operator(PauliTerm("X3"), 1)

    def test_get_sparse_operator_n_qubits_not_specified(self):
        expected = csc_matrix(
            ([1, 1, 1, 1], ([1, 0, 3, 2], [0, 1, 2, 3])), shape=(4, 4)
        )
        # Test PauliTerm.
        self.assertTrue(
            numpy.allclose(get_sparse_operator(PauliTerm("X1")).A, expected.A)
        )
        # Test PauliSum.
        self.assertTrue(
            numpy.allclose(
                get_sparse_operator(PauliSum([PauliTerm("X1")])).A, expected.A
            )
        )

    def test_get_sparse_operator_with_different_qubit_order(self):
        op = PauliTerm("(2+0j)*Z0*Z3")
        op2 = PauliTerm("(2+0j)*Z3*Z0")
        self.assertTrue(
            numpy.allclose(get_sparse_operator(op).A, get_sparse_operator(op2).A)
        )


class ExpectationTest(unittest.TestCase):
    def test_expectation_correct_sparse_matrix(self):
        operator = get_sparse_operator(PauliTerm("X0"), n_qubits=2)
        vector = numpy.array([0.0, 1.0j, 0.0, 1.0j])
        self.assertAlmostEqual(expectation(operator, vector), 2.0)

        density_matrix = scipy.sparse.csc_matrix(
            numpy.outer(vector, numpy.conjugate(vector))
        )
        self.assertAlmostEqual(expectation(operator, density_matrix), 2.0)

    def test_expectation_handles_column_vector(self):
        operator = get_sparse_operator(PauliTerm("X0"), n_qubits=2)
        vector = numpy.array([[0.0], [1.0j], [0.0], [1.0j]])
        self.assertAlmostEqual(expectation(operator, vector), 2.0)

    def test_expectation_correct_zero(self):
        operator = get_sparse_operator(PauliTerm("X0"), n_qubits=2)
        vector = numpy.array([1j, -1j, -1j, -1j])
        self.assertAlmostEqual(expectation(operator, vector), 0.0)



################################################################################
# © Copyright 2021-2022 Zapata Computing Inc.
################################################################################
"""Definition of predefined gate matrices and related utility functions."""
import numpy as np
import sympy

# --- non-parametric gates ---


def x_matrix():
    return sympy.Matrix([[0, 1], [1, 0]])


def y_matrix():
    return sympy.Matrix([[0, -1j], [1j, 0]])


def z_matrix():
    return sympy.Matrix([[1, 0], [0, -1]])


def h_matrix():
    return sympy.Matrix(
        [
            [(1 / np.sqrt(2)), (1 / np.sqrt(2))],
            [(1 / np.sqrt(2)), (-1 / np.sqrt(2))],
        ]
    )


def i_matrix():
    return sympy.Matrix([[1, 0], [0, 1]])


def s_matrix():
    return sympy.Matrix(
        [
            [1, 0],
            [0, 1j],
        ]
    )


def t_matrix():
    return sympy.Matrix(
        [
            [1, 0],
            [0, sympy.exp(1j * np.pi / 4)],
        ]
    )


# --- gates with a single param ---


def rx_matrix(angle):
    return sympy.Matrix(
        [
            [sympy.cos(angle / 2), -1j * sympy.sin(angle / 2)],
            [-1j * sympy.sin(angle / 2), sympy.cos(angle / 2)],
        ]
    )


def ry_matrix(angle):
    return sympy.Matrix(
        [
            [
                sympy.cos(angle / 2),
                -1 * sympy.sin(angle / 2),
            ],
            [
                sympy.sin(angle / 2),
                sympy.cos(angle / 2),
            ],
        ]
    )


def rz_matrix(angle):
    return sympy.Matrix(
        [
            [
                sympy.exp(-1 * sympy.I * angle / 2),
                0,
            ],
            [
                0,
                sympy.exp(sympy.I * angle / 2),
            ],
        ]
    )


def rh_matrix(angle):
    phase_factor = sympy.cos(angle / 2) + 1j * sympy.sin(angle / 2)
    return phase_factor * sympy.Matrix(
        [
            [
                sympy.cos(angle / 2) - 1j / sympy.sqrt(2) * sympy.sin(angle / 2),
                -1j / sympy.sqrt(2) * sympy.sin(angle / 2),
            ],
            [
                -1j / sympy.sqrt(2) * sympy.sin(angle / 2),
                sympy.cos(angle / 2) + 1j / sympy.sqrt(2) * sympy.sin(angle / 2),
            ],
        ]
    )


def phase_matrix(angle):
    return sympy.Matrix(
        [
            [1, 0],
            [0, sympy.exp(1j * angle)],
        ]
    )


def u3_matrix(theta, phi, lambda_):
    """Based on
    https://qiskit.org/documentation/stubs/qiskit.circuit.library.U3Gate.html
    Please note that this formulation introduces a global phase, thus the division
    after the sequence of gates.
    """
    return sympy.simplify(
        (rz_matrix(phi) * ry_matrix(theta) * rz_matrix(lambda_))
        / sympy.exp(-0.5j * (phi + lambda_))
    )


def gpi_matrix(theta):
    """Based on https://ionq.com/docs/getting-started-with-native-gates"""
    return sympy.Matrix(
        [
            [0, sympy.exp(-1j * theta)],
            [sympy.exp(1j * theta), 0],
        ]
    )


def gpi2_matrix(theta):
    """Based on https://ionq.com/docs/getting-started-with-native-gates"""
    return (2 ** (-0.5)) * sympy.Matrix(
        [
            [1, -1j * sympy.exp(-1j * theta)],
            [-1j * sympy.exp(1j * theta), 1],
        ]
    )


# --- non-parametric two qubit gates ---


def cnot_matrix():
    return sympy.Matrix(
        [
            [1, 0, 0, 0],
            [0, 1, 0, 0],
            [0, 0, 0, 1],
            [0, 0, 1, 0],
        ]
    )


def cz_matrix():
    return sympy.Matrix(
        [
            [1, 0, 0, 0],
            [0, 1, 0, 0],
            [0, 0, 1, 0],
            [0, 0, 0, -1],
        ]
    )


def swap_matrix():
    return sympy.Matrix([[1, 0, 0, 0], [0, 0, 1, 0], [0, 1, 0, 0], [0, 0, 0, 1]])


def iswap_matrix():
    return sympy.Matrix([[1, 0, 0, 0], [0, 0, 1j, 0], [0, 1j, 0, 0], [0, 0, 0, 1]])


# --- parametric two qubit gates ---


def cphase_matrix(angle):
    return sympy.Matrix(
        [
            [1, 0, 0, 0],
            [0, 1, 0, 0],
            [0, 0, 1, 0],
            [0, 0, 0, sympy.exp(1j * angle)],
        ]
    )


def xx_matrix(angle):
    return sympy.Matrix(
        [
            [sympy.cos(angle / 2), 0, 0, -1j * sympy.sin(angle / 2)],
            [0, sympy.cos(angle / 2), -1j * sympy.sin(angle / 2), 0],
            [0, -1j * sympy.sin(angle / 2), sympy.cos(angle / 2), 0],
            [-1j * sympy.sin(angle / 2), 0, 0, sympy.cos(angle / 2)],
        ]
    )


def yy_matrix(angle):
    return sympy.Matrix(
        [
            [sympy.cos(angle / 2), 0, 0, 1j * sympy.sin(angle / 2)],
            [0, sympy.cos(angle / 2), -1j * sympy.sin(angle / 2), 0],
            [0, -1j * sympy.sin(angle / 2), sympy.cos(angle / 2), 0],
            [1j * sympy.sin(angle / 2), 0, 0, sympy.cos(angle / 2)],
        ]
    )


def zz_matrix(angle):
    return sympy.Matrix(
        [
            [sympy.cos(angle / 2) - 1j * sympy.sin(angle / 2), 0, 0, 0],
            [0, sympy.cos(angle / 2) + 1j * sympy.sin(angle / 2), 0, 0],
            [0, 0, sympy.cos(angle / 2) + 1j * sympy.sin(angle / 2), 0],
            [0, 0, 0, sympy.cos(angle / 2) - 1j * sympy.sin(angle / 2)],
        ]
    )


def xy_matrix(angle):
    return sympy.Matrix(
        [
            [1, 0, 0, 0],
            [0, sympy.cos(angle / 2), 1j * sympy.sin(angle / 2), 0],
            [0, 1j * sympy.sin(angle / 2), sympy.cos(angle / 2), 0],
            [0, 0, 0, 1],
        ]
    )


def ms_matrix(phi_0, phi_1):
    """Based on https://ionq.com/docs/getting-started-with-native-gates"""
    return (2 ** (-0.5)) * sympy.Matrix(
        [
            [1, 0, 0, -1j * sympy.exp(-1j * (phi_0 + phi_1))],
            [0, 1, -1j * sympy.exp(-1j * (phi_0 - phi_1)), 0],
            [0, -1j * sympy.exp(1j * (phi_0 - phi_1)), 1, 0],
            [-1j * sympy.exp(1j * (phi_0 + phi_1)), 0, 0, 1],
        ]
    )


def delay_matrix(_duration):
    # Note that _duration parameter is not used when constructing the
    # matrix, because Delay always acts as identity.
    # Therefore, the Delay gate will always evaluate to identity on simulator,
    # but it will have an effect if used on an actual hardware.
    return i_matrix()



################################################################################
# © Copyright 2021-2022 Zapata Computing Inc.
################################################################################
from typing import Callable, Union

from . import _gates, _matrices

# GatePrototype is a function that accepts *args of type `_gates.Parameter` and returns
# a `_gates.Gate`.
GatePrototype = Callable[..., _gates.Gate]
GateRef = Union[_gates.Gate, GatePrototype]


def make_parametric_gate_prototype(
    name, matrix_factory, num_qubits, is_hermitian=False
) -> GatePrototype:
    def _factory(*gate_parameters: _gates.Parameter):
        # TODO: check if len(gate_parameters) == len(arguments of matrix_factory)
        return _gates.MatrixFactoryGate(
            name, matrix_factory, gate_parameters, num_qubits, is_hermitian
        )

    return _factory


def builtin_gate_by_name(name) -> GateRef:
    """Retrieves gate factory by its name.

    Raises:
        KeyError: when there's no built-in gate for this name.
    """
    return globals()[name]


# --- non-parametric, single qubit gates ---

X = _gates.MatrixFactoryGate("X", _matrices.x_matrix, (), 1, is_hermitian=True)
Y = _gates.MatrixFactoryGate("Y", _matrices.y_matrix, (), 1, is_hermitian=True)
Z = _gates.MatrixFactoryGate("Z", _matrices.z_matrix, (), 1, is_hermitian=True)
H = _gates.MatrixFactoryGate("H", _matrices.h_matrix, (), 1, is_hermitian=True)
I = _gates.MatrixFactoryGate(  # noqa: E741
    "I", _matrices.i_matrix, (), 1, is_hermitian=True
)
S = _gates.MatrixFactoryGate("S", _matrices.s_matrix, (), 1)
T = _gates.MatrixFactoryGate("T", _matrices.t_matrix, (), 1)


# --- parametric, single qubit gates ---


RX = make_parametric_gate_prototype("RX", _matrices.rx_matrix, 1)
RY = make_parametric_gate_prototype("RY", _matrices.ry_matrix, 1)
RZ = make_parametric_gate_prototype("RZ", _matrices.rz_matrix, 1)
RH = make_parametric_gate_prototype("RH", _matrices.rh_matrix, 1)
PHASE = make_parametric_gate_prototype("PHASE", _matrices.phase_matrix, 1)
U3 = make_parametric_gate_prototype("U3", _matrices.u3_matrix, 1)
GPi = make_parametric_gate_prototype("GPi", _matrices.gpi_matrix, 1, is_hermitian=True)
GPi2 = make_parametric_gate_prototype("GPi2", _matrices.gpi2_matrix, 1)


# --- non-parametric, two qubit gates ---

CNOT = _gates.MatrixFactoryGate("CNOT", _matrices.cnot_matrix, (), 2, is_hermitian=True)
CZ = _gates.MatrixFactoryGate("CZ", _matrices.cz_matrix, (), 2, is_hermitian=True)
SWAP = _gates.MatrixFactoryGate("SWAP", _matrices.swap_matrix, (), 2, is_hermitian=True)
ISWAP = _gates.MatrixFactoryGate("ISWAP", _matrices.iswap_matrix, (), 2)


# --- parametric, two qubit gates ---

CPHASE = make_parametric_gate_prototype("CPHASE", _matrices.cphase_matrix, 2)
XX = make_parametric_gate_prototype("XX", _matrices.xx_matrix, 2)
YY = make_parametric_gate_prototype("YY", _matrices.yy_matrix, 2)
ZZ = make_parametric_gate_prototype("ZZ", _matrices.zz_matrix, 2)
XY = make_parametric_gate_prototype("XY", _matrices.xy_matrix, 2)
MS = make_parametric_gate_prototype("MS", _matrices.ms_matrix, 2)


# --- misc ----
# The Delay gate acts as an identity, however delays further circuit execution
# on the given qubit by n-units of time.
Delay = make_parametric_gate_prototype(
    "Delay", _matrices.delay_matrix, 1, is_hermitian=True
)



################################################################################
# © Copyright 2022 Zapata Computing Inc.
################################################################################
from .circuit_runner import BaseCircuitRunner, CircuitRunner
from .estimation import EstimationTask
from .wavefunction_simulator import BaseWavefunctionSimulator, WavefunctionSimulator



################################################################################
# © Copyright 2021-2022 Zapata Computing Inc.
################################################################################

import os
from collections import Counter
from math import sqrt

import numpy as np
import pytest
from sympy import I, Matrix, Symbol, cos, exp, sin

from orquestra.quantum.circuits._builtin_gates import RX, RY, U3, H, X
from orquestra.quantum.circuits._circuit import Circuit
from orquestra.quantum.runners import SymbolicSimulator
from orquestra.quantum.testing import create_random_wavefunction
from orquestra.quantum.utils import RNDSEED, bitstring_to_tuple
from orquestra.quantum.wavefunction import (
    Wavefunction,
    load_wavefunction,
    sample_from_wavefunction,
    save_wavefunction,
)


def remove_file_if_exists(filename):
    try:
        os.remove(filename)
    except OSError:
        pass


class TestInitSystemInZeroState:
    def test_init_system_returns_numpy_array(self):
        wf = Wavefunction.zero_state(2)
        assert isinstance(wf._amplitude_vector, np.ndarray)

    def test_constructor_returns_numpy_array_for_no_symbols(self):
        wf = Wavefunction([1.0, 0, 0, 0])
        assert isinstance(wf._amplitude_vector, np.ndarray)

    def test_constructor_returns_sympy_matrix_for_free_symbols(self):
        wf = Wavefunction([0.25, 0, Symbol("alpha"), 0])
        assert isinstance(wf._amplitude_vector, Matrix)

    @pytest.mark.parametrize(
        "input_list", [[], np.zeros(17), create_random_wavefunction(3)[:-1]]
    )
    def test_init_fails_when_len_of_passed_list_is_not_power_of_two(self, input_list):
        with pytest.raises(ValueError):
            Wavefunction(input_list)

    @pytest.mark.parametrize("input_list", [np.ones(8), np.zeros(4)])
    def test_init_fails_when_passed_list_has_no_free_symbols_and_no_unity(
        self, input_list
    ):
        with pytest.raises(ValueError):
            Wavefunction(input_list)

    @pytest.mark.parametrize(
        "input_list",
        [
            [Symbol("alpha"), 2.0],
            [
                Symbol("alpha"),
                Symbol("beta"),
                sqrt(3) / 2,
                sqrt(3) / 2,
            ],
        ],
    )
    def test_init_fails_when_passed_list_has_free_symbols_and_exceeds_unity(
        self, input_list
    ):
        with pytest.raises(ValueError):
            Wavefunction(input_list)

    @pytest.mark.parametrize("n_qubits", [1, 2, 3, 4, 5])
    def test_init_system_returns_expected_wavefunction_size(self, n_qubits):
        wavefunction = Wavefunction.zero_state(n_qubits=n_qubits)

        # Check length
        assert len(wavefunction) == 2**n_qubits

        # Check internal property
        assert wavefunction.n_qubits == n_qubits

        # Check amplitude of zero state
        assert wavefunction[0] == 1.0

        # Check amplitude of the rest of the states
        assert not np.any(wavefunction[1:])

    def test_init_system_raises_warning_for_non_ints(self):
        with pytest.warns(UserWarning):
            Wavefunction.zero_state(1.234)

    @pytest.mark.parametrize("n_qubits", [0, -1, -2])
    def test_init_system_fails_on_invalid_params(self, n_qubits):
        with pytest.raises(ValueError):
            Wavefunction.zero_state(n_qubits=n_qubits)


class TestInitSystemInDickeState:
    @pytest.mark.parametrize("num_qubits", [-1, 0])
    def test_function_fails_for_invalid_number_of_qubits(self, num_qubits):
        with pytest.raises(ValueError):
            Wavefunction.dicke_state(num_qubits, 2)

    @pytest.mark.parametrize("hamming_weight", [-1, 3, 3.36])
    def test_function_fails_for_invalid_hamming_weight(self, hamming_weight):
        with pytest.raises(ValueError):
            Wavefunction.dicke_state(2, hamming_weight)

    @pytest.mark.parametrize(
        "expected_set_states,expected_amplitude,hamming_weight",
        [
            ([0], 1.0, 0),
            ([1, 2, 4, 8], 1 / np.sqrt(4), 1),
            ([3, 5, 6, 9, 10, 12], 1 / np.sqrt(6), 2),
            ([7, 11, 13, 14], 1 / np.sqrt(4), 3),
            ([int("1111", base=2)], 1.0, 4),
        ],
    )
    def test_function_returns_expected_wf_for_given_hamming_weight(
        self, expected_set_states, expected_amplitude, hamming_weight
    ):
        wf = Wavefunction.dicke_state(4, hamming_weight=hamming_weight)

        unique = np.unique(wf)
        unique = np.delete(unique, np.where(unique == 0.0))
        assert len(unique) == 1
        assert unique.item() == expected_amplitude

        indices = np.where(wf.amplitudes == unique.item())[0]
        np.testing.assert_array_equal(np.array(expected_set_states), indices)


class TestFunctions:
    @pytest.fixture
    def symbolic_wf(self) -> Wavefunction:
        return Wavefunction([Symbol("alpha"), 0.5, Symbol("beta"), 0.5])

    @pytest.fixture
    def numeric_wf(self) -> Wavefunction:
        return Wavefunction([0.5, 0.5, 0.5, 0.5])

    @pytest.mark.parametrize("new_val", [1.0, -1.0])
    def test_set_item_raises_error_for_invalid_sets(self, symbolic_wf, new_val):
        with pytest.raises(ValueError):
            symbolic_wf[0] = new_val

    @pytest.mark.parametrize("new_value", [0.5, Symbol("gamma")])
    def test_set_item_passes_if_still_below_unity(self, symbolic_wf, new_value):
        symbolic_wf[0] = new_value

        assert symbolic_wf[0] == new_value

    def test_iterator(self, symbolic_wf):
        for i, elem in enumerate(symbolic_wf):
            assert elem == symbolic_wf[i]

    @pytest.mark.parametrize(
        "symbol_map", [{"alpha": 1.0}, {"alpha": 0.5, "beta": 0.6}]
    )
    def test_bindings_fail_like_setitem(self, symbolic_wf, symbol_map):
        with pytest.raises(ValueError):
            symbolic_wf.bind(symbol_map)

    def test_bind_returns_new_object_for_symbolic_wf(self, symbolic_wf):
        assert symbolic_wf is not symbolic_wf.bind({})

    def test_bind_does_not_return_new_object_for_numeric_wf(self, numeric_wf):
        assert numeric_wf is numeric_wf.bind({})

    def test_binding_all_symbols_returns_numpy_array(self, symbolic_wf: Wavefunction):
        assert isinstance(
            symbolic_wf.bind({"alpha": 0.5, "beta": 0.5})._amplitude_vector, np.ndarray
        )

    @pytest.mark.parametrize("other_obj", [[], np.zeros(8)])
    def test_eq_returns_false_for_non_wavefunction_objects(
        self, symbolic_wf, numeric_wf, other_obj
    ):
        assert not (symbolic_wf == other_obj)
        assert not (numeric_wf == other_obj)

    def test_eq_returns_true_for_objects_with_equal_wavefunctions(
        self, symbolic_wf: Wavefunction, numeric_wf: Wavefunction
    ):
        test_wf = Wavefunction(symbolic_wf._amplitude_vector)
        assert symbolic_wf == test_wf

        test_wf = Wavefunction(numeric_wf._amplitude_vector)
        assert numeric_wf == test_wf


class TestRepresentations:
    def test_string_output_of_symbolic_wavefunction(self):
        wf = Wavefunction([Symbol("alpha"), 0])

        wf_str = wf.__str__()

        assert "alpha" in wf_str
        assert wf_str.endswith("])")
        assert wf_str.startswith("Wavefunction([")

    def test_string_output_of_numeric_wavefunction(self):
        wf = Wavefunction([1j, 0])

        wf_str = wf.__str__()

        assert "j" in wf_str
        assert wf_str.endswith("])")
        assert wf_str.startswith("Wavefunction([")

    @pytest.mark.parametrize(
        "wf", [Wavefunction.zero_state(2), Wavefunction([Symbol("alpha"), 0.0])]
    )
    def test_amplitudes_and_probs_output_type(self, wf: Wavefunction):
        if len(wf.free_symbols) > 0:
            assert wf.amplitudes.dtype == object
            assert wf.get_probabilities().dtype == object
        else:
            assert wf.amplitudes.dtype == np.complex128
            assert wf.get_probabilities().dtype == np.float64

    @pytest.mark.parametrize(
        "wf_vec",
        [
            [1.0, 0.0],
            [0.5, 0.5, 0.5, 0.5],
            [1 / sqrt(2), 0, 0, 0, 0, 0, 0, 1 / sqrt(2)],
        ],
    )
    def test_get_outcome_probs(self, wf_vec):
        wf = Wavefunction(wf_vec)
        probs_dict = wf.get_outcome_probs()

        assert all([len(key) == wf.n_qubits for key in probs_dict.keys()])

        for key in probs_dict.keys():
            assert len(key) == wf.n_qubits

            assert wf.get_probabilities()[int(key, 2)] == probs_dict[key]


class TestGates:
    @pytest.fixture
    def simulator(self) -> SymbolicSimulator:
        return SymbolicSimulator()

    @pytest.mark.parametrize(
        "circuit, expected_wavefunction",
        [
            (
                Circuit([RX(Symbol("theta"))(0)]),
                Wavefunction(
                    [1.0 * cos(Symbol("theta") / 2), -1j * sin(Symbol("theta") / 2)]
                ),
            ),
            (
                Circuit([X(0), RY(Symbol("theta"))(0)]),
                Wavefunction(
                    [
                        -1.0 * sin(Symbol("theta") / 2),
                        1.0 * cos(Symbol("theta") / 2),
                    ]
                ),
            ),
            (
                Circuit(
                    [H(0), U3(Symbol("theta"), Symbol("phi"), Symbol("lambda"))(0)]
                ),
                Wavefunction(
                    [
                        cos(Symbol("theta") / 2) / sqrt(2)
                        + -exp(I * Symbol("lambda"))
                        * sin(Symbol("theta") / 2)
                        / sqrt(2),
                        exp(I * Symbol("phi")) * sin(Symbol("theta") / 2) / sqrt(2)
                        + exp(I * (Symbol("lambda") + Symbol("phi")))
                        * cos(Symbol("theta") / 2)
                        / sqrt(2),
                    ]
                ),
            ),
        ],
    )
    def test_wavefunction_works_as_expected_with_symbolic_circuits(
        self,
        simulator: SymbolicSimulator,
        circuit: Circuit,
        expected_wavefunction: Wavefunction,
    ):
        returned_wavefunction = simulator.get_wavefunction(circuit)

        assert returned_wavefunction == expected_wavefunction


def test_real_wavefunction_io():
    wf = Wavefunction([0, 1, 0, 0, 0, 0, 0, 0])
    save_wavefunction(wf, "wavefunction.json")
    loaded_wf = load_wavefunction("wavefunction.json")
    assert np.allclose(wf.amplitudes, loaded_wf.amplitudes)
    remove_file_if_exists("wavefunction.json")


def test_imag_wavefunction_io():
    wf = Wavefunction([0, 1j, 0, 0, 0, 0, 0, 0])
    save_wavefunction(wf, "wavefunction.json")
    loaded_wf = load_wavefunction("wavefunction.json")
    assert np.allclose(wf.amplitudes, loaded_wf.amplitudes)
    remove_file_if_exists("wavefunction.json")


def test_sample_from_wavefunction():
    wavefunction = create_random_wavefunction(4, seed=RNDSEED)
    num_samples = 100000

    samples = sample_from_wavefunction(wavefunction, num_samples, seed=RNDSEED)
    sampled_dict = Counter(samples)

    sampled_probabilities = []
    for num in range(len(wavefunction)):
        bitstring = format(num, "b")
        while len(bitstring) < wavefunction.n_qubits:
            bitstring = "0" + bitstring
        measurement = bitstring_to_tuple(bitstring)[::-1]
        sampled_probabilities.append(sampled_dict[measurement] / num_samples)

    probabilities = wavefunction.get_probabilities()
    for sampled_prob, exact_prob in zip(sampled_probabilities, probabilities):
        assert np.allclose(sampled_prob, exact_prob, atol=0.01)


def test_sample_from_wavefunction_column_vector():
    n_qubits = 4
    expected_bitstring = (0, 0, 0, 1)
    amplitudes = np.array([0] * (2**n_qubits)).reshape(2**n_qubits, 1)
    amplitudes[1] = 1  # |0001> will be measured in all cases.
    wavefunction = Wavefunction(amplitudes)
    sample = set(sample_from_wavefunction(wavefunction, 500))
    assert len(sample) == 1
    assert sample.pop() == expected_bitstring


def test_sample_from_wavefunction_row_vector():
    n_qubits = 4
    expected_bitstring = (0, 0, 0, 1)
    amplitudes = np.array([0] * (2**n_qubits))
    amplitudes[1] = 1  # |0001> will be measured in all cases.
    wavefunction = Wavefunction(amplitudes)
    sample = set(sample_from_wavefunction(wavefunction, 500))
    assert len(sample) == 1
    assert sample.pop() == expected_bitstring


def test_sample_from_wavefunction_list():
    n_qubits = 4
    expected_bitstring = (0, 0, 0, 1)
    amplitudes = [0] * (2**n_qubits)
    amplitudes[1] = 1  # |0001> will be measured in all cases.
    wavefunction = Wavefunction(amplitudes)
    sample = set(sample_from_wavefunction(wavefunction, 500))
    assert len(sample) == 1
    assert sample.pop() == expected_bitstring


@pytest.mark.parametrize("n_samples", [-1, 0])
def test_sample_from_wavefunction_fails_for_invalid_n_samples(n_samples):
    n_qubits = 4
    amplitudes = [0] * (2**n_qubits)
    amplitudes[1] = 1
    wavefunction = Wavefunction(amplitudes)
    with pytest.raises(ValueError):
        sample_from_wavefunction(wavefunction, n_samples)



################################################################################
# © Copyright 2022 Zapata Computing Inc.
################################################################################
import copy
import json
import math
import sys
import warnings
from itertools import product
from typing import Any, Callable, Dict, List, Tuple, Union

import numpy as np

from ..typing import AnyPath


class MeasurementOutcomeDistribution:
    """A probability distribution defined on discrete non-negative integer
    sequences. Normalization is performed by default, unless otherwise specified.

    Args:
        input_dict:  dictionary representing the probability distribution where
            the keys are non-negative integer sequences represented as tuples
            and the values are non-negative floats.
        normalize: boolean variable specifying whether the input_dict gets
            normalized or not.
    Attributes:
        distribution_dict: dictionary representing the probability
            distribution where the keys are non-negative integer sequences
            represented as tuples and the values are non-negative floats.
    """

    def __init__(
        self,
        input_dict: Dict[Union[str, Tuple[int, ...]], float],
        normalize: bool = True,
    ):
        # First check if we are initializing from binary strings
        preprocessed_input_dict = preprocess_distibution_dict(input_dict)

        if is_measurement_outcome_distribution(
            preprocessed_input_dict
        ):  # accept the input dict only if it is a prob distribution
            if is_normalized(preprocessed_input_dict):
                self.distribution_dict = preprocessed_input_dict
            else:
                if normalize:
                    self.distribution_dict = normalize_measurement_outcome_distribution(
                        preprocessed_input_dict
                    )
                else:
                    warnings.warn(
                        "MeasurementOutcomeDistribution object is not normalized."
                    )
                    self.distribution_dict = preprocessed_input_dict
        else:
            raise RuntimeError(
                "Initialization of MeasurementOutcomeDistribution object FAILED: "
                "the input dictionary is not a non-negative integer sequence "
                "probability distribution. Check keys (same-length non-negative integer"
                " tuples) and values (non-negative floats)."
            )

    def __repr__(self) -> str:
        output = f"MeasurementOutcomeDistribution(input={self.distribution_dict})"
        return output

    def get_number_of_subsystems(self) -> int:
        """Compute how many subsystems the measurement outcome is composed of.
        This corresponds to the number of qubits in a digital quantum computer.

        Returns:
            int: number of subsystems in the measurement outcome
                (i.e. non-negative integer sequence length).
        """
        return len(
            list(self.distribution_dict.keys())[0]
        )  # already checked in __init__ that all keys have the same length

    def subdistribution(
        self, active_qubits: List[int]
    ) -> "MeasurementOutcomeDistribution":
        """
        Returns a subdistribution, which represents measurements outcome including only
        the specified qubits.

        Returns:
            active_qubits: indices of the qubits that should be included in
                the subdistribution. Order of the indices will be preserved in
                the order of the result bitstrings.

        """
        new_counts: Dict[Union[str, Tuple[int, ...]], float] = {}

        # check for no out of range indexes
        if max(active_qubits) + 1 > len(list(self.distribution_dict.keys())[0]):
            raise ValueError("Active qubit indeces contains out of bound index.")

        # check for duplicate indexes
        if len(active_qubits) != len(set(active_qubits)):
            raise ValueError("There exist duplicate indices in the active qubit list")

        for key in copy.deepcopy(list(self.distribution_dict.keys())):
            new_key = "".join(str(key[i]) for i in active_qubits)
            new_counts[new_key] = self.distribution_dict.pop(key) + new_counts.get(
                new_key, 0
            )
        normalize = is_normalized(self.distribution_dict)
        return MeasurementOutcomeDistribution(new_counts, normalize=normalize)


def preprocess_distibution_dict(
    input_dict: Dict[Union[str, Tuple[int, ...]], float]
) -> Dict[Union[str, Tuple[int, ...]], float]:
    res_dict: Dict[Union[str, Tuple[int, ...]], float] = {}
    for key, value in input_dict.items():
        if isinstance(key, str):
            res_dict[tuple(map(int, key if "," not in key else key.split(",")))] = value
        elif isinstance(key, tuple):
            res_dict[key] = value
        else:
            raise RuntimeError(
                "Initialization of MeasurementOutcomeDistribution object FAILED:"
                "The non-tuple dictionary keys are not consistent. "
                "Check that the keys are either same-length tuples, binary strings "
                "or comma-separated non-negative integer strings."
            )

    return res_dict


def _is_non_negative(input_dict: Dict[Union[str, Tuple[int, ...]], float]) -> bool:
    """Check if the input dictionary values are non-negative.

    Args:
        input_dict (dict): dictionary.

    Returns:
        bool: boolean variable indicating whether dict values are non negative or not.
    """
    return all(value >= 0 for value in input_dict.values())


def _is_key_length_fixed(input_dict: Dict[Union[str, Tuple[int, ...]], float]) -> bool:
    """Check if the input dictionary keys are same-length.

    Args:
        input_dict (dict): dictionary.

    Returns:
        bool: boolean variable indicating whether dict keys are same-length or not.
    """
    key_length = len(list(input_dict.keys())[0])
    return all(len(key) == key_length for key in input_dict.keys())


def _are_keys_non_negative_integer_tuples(
    input_dict: Dict[Union[str, Tuple[int, ...]], float]
) -> bool:
    """Check if the input dictionary keys are tuples containing non-negative integers.

    Args:
        input_dict (dict): dictionary.

    Returns:
        bool: boolean variable indicating whether dict keys are binary strings or not.
    """
    return all(
        all(isinstance(sub, (int, np.integer)) and sub >= 0 for sub in key)
        for key in input_dict.keys()
    )


def is_measurement_outcome_distribution(
    input_dict: Dict[Union[str, Tuple[int, ...]], float]
) -> bool:
    """Check if the input dictionary is a non-negative integer sequence distribution, i.e.:
            - keys are same-length tuples of non-negative integers,
            - values are non negative.

    Args:
        input_dict: dictionary representing the probability distribution where
            the keys are non-negative integer sequences represented as tuples
            and the values are floats.

    Returns:
        Boolean variable indicating whether the non-negative integer sequence
            distribution is well defined or not.
    """
    return (
        (not input_dict == {})
        and _is_non_negative(input_dict)
        and _is_key_length_fixed(input_dict)
        and _are_keys_non_negative_integer_tuples(input_dict)
    )


def is_normalized(input_dict: Dict[Union[str, Tuple[int, ...]], float]) -> bool:
    """Check if a measurement outcome distribution is normalized.

    Args:
        input_dict: dictionary representing the probability distribution
            where the keys are non-negative integer sequences represented
            as tuples and the values are floats.

    Returns:
        Boolean value indicating whether the measurement outcome distribution
            is normalized.
    """
    norm = sum(input_dict.values())
    return math.isclose(norm, 1)


def normalize_measurement_outcome_distribution(
    measurement_outcome_distribution: Dict[Union[str, Tuple[int, ...]], float]
) -> Dict:
    """Normalize a measurement outcome distribution.

    Args:
        measurement_outcome_distribution: dictionary representing the probability
            distribution where the keys are non-negative integer sequences represented
            as tuples and the values are floats.

    Returns:
        Dictionary representing the normalized probability distribution where the keys
            are non-negative integer sequences represented as tuples and the values
            are floats.
    """
    norm = sum(measurement_outcome_distribution.values())
    if norm == 0:
        raise ValueError(
            "Normalization of MeasurementOutcomeDistribution FAILED:"
            " input dict is empty (all zero values)."
        )
    elif 0 < norm < sys.float_info.min:
        raise ValueError(
            "Normalization of MeasurementOutcomeDistribution FAILED: too small values."
        )
    elif norm == 1:
        return measurement_outcome_distribution
    else:
        for key in measurement_outcome_distribution:
            measurement_outcome_distribution[key] *= 1.0 / norm
        return measurement_outcome_distribution


def change_tuple_dict_keys_to_comma_separated_integers(dict):
    return {
        ",".join(map(str, key)) if isinstance(key, tuple) else key: value
        for key, value in dict.items()
    }


def save_measurement_outcome_distribution(
    distribution: MeasurementOutcomeDistribution, filename: AnyPath
) -> None:
    """Save a measurement outcome distribution to a file.

    Args:
        distribution (MeasurementOutcomeDistribution): the measurement outcome
        distribution file (str or file-like object): the name of the file,
            or a file-like object
    """
    dictionary: Dict[str, Any] = {}

    # Need to convert tuples to strings for JSON encoding
    preprocessed_distribution_dict = change_tuple_dict_keys_to_comma_separated_integers(
        distribution.distribution_dict
    )

    dictionary["measurement_outcome_distribution"] = preprocessed_distribution_dict
    with open(filename, "w") as f:
        f.write(json.dumps(dictionary, indent=2))


def save_measurement_outcome_distributions(
    measurement_outcome_distribution: List[MeasurementOutcomeDistribution],
    filename: str,
) -> None:
    """Save a set of measurement outcome distributions to a file.

    Args:
       measurement_outcome_distribution (list): a list of distributions to be saved
       file (str): the name of the file
    """
    dictionary: Dict[str, Any] = {}
    dictionary["measurement_outcome_distribution"] = []

    for distribution in measurement_outcome_distribution:
        dictionary["measurement_outcome_distribution"].append(
            change_tuple_dict_keys_to_comma_separated_integers(
                distribution.distribution_dict
            )
        )

    with open(filename, "w") as f:
        f.write(json.dumps(dictionary, indent=2))


def load_measurement_outcome_distribution(file: str) -> MeasurementOutcomeDistribution:
    """Load a measurement outcome distribution from a json file.

    Arguments:
        file (str): the name of the file

    Returns:
        object: a python object loaded from the measurement_outcome_distribution
    """
    if isinstance(file, str):
        with open(file, "r") as f:
            data = json.load(f)
    else:
        data = json.load(file)

    try:
        distribution = MeasurementOutcomeDistribution(data["bitstring_distribution"])
    except KeyError:
        distribution = MeasurementOutcomeDistribution(
            data["measurement_outcome_distribution"]
        )
    return distribution


def load_measurement_outcome_distributions(
    file: str,
) -> List[MeasurementOutcomeDistribution]:
    """Load a list of measurement_outcome_distribution from a json file.

    Arguments:
        file: the name of the file.

    Returns:
        A list of measurement outcome distributions loaded
         from the measurement_outcome_distribution
    """
    if isinstance(file, str):
        with open(file, "r") as f:
            data = json.load(f)
    else:
        data = json.load(file)

    distribution_list = []
    for i in range(len(data["measurement_outcome_distribution"])):
        try:
            distribution_list.append(
                MeasurementOutcomeDistribution(data["bitstring_distribution"][i])
            )
        except KeyError:
            distribution_list.append(
                MeasurementOutcomeDistribution(
                    data["measurement_outcome_distribution"][i]
                )
            )

    return distribution_list


def create_bitstring_distribution_from_probability_distribution(
    prob_distribution: np.ndarray,
) -> MeasurementOutcomeDistribution:
    """Create a well defined bitstring distribution starting from a probability
    distribution.

    Args:
        probability distribution: The probabilities of the various states in the
            wavefunction.

    Returns:
        The MeasurementOutcomeDistribution object corresponding to
            the input measurements.
    """
    # Create dictionary of bitstring tuples as keys with probability as value
    keys = product([0, 1], repeat=int(np.log2(len(prob_distribution))))
    prob_dict: Dict[Union[str, Tuple[int, ...]], float] = {
        key: float(value) for key, value in zip(keys, prob_distribution)
    }

    return MeasurementOutcomeDistribution(prob_dict)


def evaluate_distribution_distance(
    target_distribution: MeasurementOutcomeDistribution,
    measured_distribution: MeasurementOutcomeDistribution,
    distance_measure_function: Callable,
    **kwargs,
) -> float:
    """Evaluate the distance between two measurement outcome distributions - the target
    distribution and the one predicted (measured) by your model - based on the given
    distance measure.

    Args:
         target_distribution: The target measurement outcome probability distribution
         measured_distribution: The measured measurement outcome probability
          distribution
         distance_measure_function: function used to calculate the distance measure
            Currently implemented: clipped negative log-likelihood, maximum mean
            discrepancy (MMD).

         Additional distance measure parameters can be passed as keyword arguments.

    Returns:
         The value of the distance measure.
    """
    # Check inputs are MeasurementOutcomeDistribution objects
    if not isinstance(
        target_distribution, MeasurementOutcomeDistribution
    ) or not isinstance(measured_distribution, MeasurementOutcomeDistribution):
        raise TypeError(
            "Arguments of evaluate_cost_function must"
            " be of type MeasurementOutcomeDistribution."
        )

    # Check inputs are defined on consistent measurement outcome domains
    if (
        target_distribution.get_number_of_subsystems()
        != measured_distribution.get_number_of_subsystems()
    ):
        raise RuntimeError(
            "Measurement Outcome Distribution Distance Evaluation FAILED: target "
            "and measured distributions are defined on tuples of different length."
        )

    # Check inputs are both normalized (or not normalized)
    if is_normalized(target_distribution.distribution_dict) != is_normalized(
        measured_distribution.distribution_dict
    ):
        raise RuntimeError(
            "Measurement Outcome Distribution Distance Evaluation FAILED: one among "
            "target and measured distribution is normalized, whereas the other is not."
        )

    return distance_measure_function(
        target_distribution, measured_distribution, **kwargs
    )



################################################################################
# © Copyright 2021-2022 Zapata Computing Inc.
################################################################################
from abc import abstractmethod
from functools import singledispatch
from numbers import Number
from typing import Dict, Iterable, Protocol, Tuple, TypeVar

import sympy

from ..typing import Parameter, ParameterizedVector

T = TypeVar("T", bound="Operation")


class Operation(Protocol):
    """Represents arbitrary operation applicable to a circuit or wavefunction."""

    @property
    @abstractmethod
    def params(self) -> Tuple[Parameter, ...]:
        """Parameters of this operation."""

    @abstractmethod
    def bind(self: T, symbols_map: Dict[sympy.Symbol, Parameter]) -> T:
        """Create new operation by replacing free symbols in operation params.

        The operation returned by this method should be of the same type
        as self, e.g. binding parameters to GateOperation should produce
        GateOperation.
        """

    @abstractmethod
    def replace_params(self: T, new_params: Tuple[Parameter, ...]) -> T:
        """Create new operation by replacing params.

        The difference between bind and replace params is that it bind performs
        parameter substitution - in particular, parameters without free symbols
        are unaffected by bind, whereas replace_params replaces *all* params.
        """

    @abstractmethod
    def apply(self: T, amplitude_vector: ParameterizedVector) -> ParameterizedVector:
        """Apply this operation to given wavefunction.

        Operations that cannot be applied deterministically should raise
        NotImplementedError.
        """

    @property
    @abstractmethod
    def free_symbols(self) -> Iterable[sympy.Symbol]:
        """Free symbols parameterizing this operation.

        Note that number of free_symbols is unrelated to number of params.
        Some params can be expressions with multiple free symbols, while other params
        might not comprise free symbols at all.
        """


@singledispatch
def sub_symbols(parameter, symbols_map: Dict[sympy.Symbol, Parameter]) -> Parameter:
    raise NotImplementedError()


@sub_symbols.register
def _sub_symbols_in_number(
    parameter: Number, symbols_map: Dict[sympy.Symbol, Parameter]
) -> Number:
    return parameter


@sub_symbols.register
def _sub_symbols_in_expression(
    parameter: sympy.Expr, symbols_map: Dict[sympy.Symbol, Parameter]
) -> sympy.Expr:
    return parameter.subs(symbols_map)


@sub_symbols.register
def _sub_symbols_in_symbol(
    parameter: sympy.Symbol, symbols_map: Dict[sympy.Symbol, Parameter]
) -> Parameter:
    return symbols_map.get(parameter, parameter)


def get_free_symbols(parameters: Tuple[Parameter, ...]) -> Iterable[sympy.Symbol]:
    symbols = set(
        symbol
        for param in parameters
        if isinstance(param, sympy.Expr)
        for symbol in param.free_symbols
    )
    return sorted(symbols, key=str)



################################################################################
# © Copyright 2021-2022 Zapata Computing Inc.
################################################################################
import json
import os
import random
from collections import Counter

import numpy as np
import pytest

from orquestra.quantum.distributions import MeasurementOutcomeDistribution
from orquestra.quantum.measurements import (
    Measurements,
    _check_sample_elimination,
    convert_bitstring_to_int,
    get_expectation_value_from_frequencies,
)
from orquestra.quantum.operators import PauliTerm
from orquestra.quantum.utils import (
    RNDSEED,
    convert_tuples_to_bitstrings,
    get_ordered_list_of_bitstrings,
)


def remove_file_if_exists(filename):
    try:
        os.remove(filename)
    except OSError:
        pass


def test_convert_bitstring_to_int():
    bitstring = (0, 1, 0, 1, 0, 1)
    assert convert_bitstring_to_int(bitstring) == 42


def test_get_expectation_value_from_frequencies():
    bitstrings = ["001", "001", "110", "000"]
    bitstring_frequencies = dict(Counter(bitstrings))
    marked_qubits = (1, 2)
    assert np.isclose(
        get_expectation_value_from_frequencies(marked_qubits, bitstring_frequencies),
        -0.5,
    )


class TestMeasurements:
    @pytest.fixture
    def bitstrings(self):
        return [
            [0, 0, 0],
            [0, 0, 1],
            [0, 1, 0],
            [0, 1, 1],
            [1, 0, 0],
            [1, 1, 0],
            [1, 1, 1],
            [1, 0, 1],
            [0, 0, 1],
        ]

    @pytest.fixture
    def counts(self):
        return {
            key: value
            for key, value in zip(
                get_ordered_list_of_bitstrings(3), [1, 2, 1, 1, 1, 1, 1, 1]
            )
        }

    @pytest.fixture
    def measurements_data(self, counts, bitstrings):
        return {
            "counts": counts,
            "bitstrings": bitstrings,
        }

    def test_io(self, measurements_data):
        # Given
        input_filename = "measurements_input_test.json"
        output_filename = "measurements_output_test.json"

        with open(input_filename, "w") as f:
            f.write(json.dumps(measurements_data, indent=2))

        # When
        measurements = Measurements.load_from_file(input_filename)
        measurements.save(output_filename)

        # Then
        with open(output_filename, "r") as f:
            output_data = json.load(f)
        assert measurements_data == output_data

        remove_file_if_exists(input_filename)
        remove_file_if_exists(output_filename)

    def test_save_for_numpy_integers(self):
        # Given
        target_bitstrings = [(0, 0, 0)]
        input_bitstrings = [(np.int8(0), np.int8(0), np.int8(0))]

        filename = "measurementstest.json"
        measurements = Measurements(input_bitstrings)
        target_measurements = Measurements(target_bitstrings)

        # When
        measurements.save(filename)

        # Then
        recreated_measurements = Measurements.load_from_file(filename)
        assert target_measurements.bitstrings == recreated_measurements.bitstrings
        remove_file_if_exists("measurementstest.json")

    def test_intialize_with_bitstrings(self):
        # Given
        bitstrings = [
            (0, 0, 0),
            (0, 0, 1),
            (0, 0, 1),
            (0, 1, 0),
            (0, 1, 1),
            (1, 0, 0),
            (1, 0, 1),
            (1, 1, 0),
            (1, 1, 1),
        ]

        # When
        measurements = Measurements(bitstrings=bitstrings)

        # Then
        assert measurements.bitstrings == [
            (0, 0, 0),
            (0, 0, 1),
            (0, 0, 1),
            (0, 1, 0),
            (0, 1, 1),
            (1, 0, 0),
            (1, 0, 1),
            (1, 1, 0),
            (1, 1, 1),
        ]

    def test_intialize_with_counts(self, counts):
        # When
        measurements = Measurements.from_counts(counts)

        # Then
        assert measurements.bitstrings == [
            (0, 0, 0),
            (0, 0, 1),
            (0, 0, 1),
            (0, 1, 0),
            (0, 1, 1),
            (1, 0, 0),
            (1, 0, 1),
            (1, 1, 0),
            (1, 1, 1),
        ]

    def test_bitstrings(self, measurements_data):
        input_filename = "measurements_input_test.json"

        with open(input_filename, "w") as f:
            f.write(json.dumps(measurements_data, indent=2))
        measurements = Measurements.load_from_file(input_filename)

        # When/Then
        assert measurements.bitstrings == [
            (0, 0, 0),
            (0, 0, 1),
            (0, 1, 0),
            (0, 1, 1),
            (1, 0, 0),
            (1, 1, 0),
            (1, 1, 1),
            (1, 0, 1),
            (0, 0, 1),
        ]

        remove_file_if_exists(input_filename)

    def test_get_counts(self, measurements_data):
        # Given
        input_filename = "measurements_input_test.json"

        with open(input_filename, "w") as f:
            f.write(json.dumps(measurements_data, indent=2))
        measurements = Measurements.load_from_file(input_filename)

        # When
        counts = measurements.get_counts()

        # Then
        assert measurements_data["counts"] == counts

        remove_file_if_exists(input_filename)

    def test_get_distribution(self, measurements_data):
        # Given
        input_filename = "measurements_input_test.json"

        with open(input_filename, "w") as f:
            f.write(json.dumps(measurements_data, indent=2))
        measurements = Measurements.load_from_file(input_filename)

        # When
        distribution = measurements.get_distribution()

        # Then
        assert distribution.distribution_dict == {
            (0, 0, 0): 1 / 9,
            (0, 0, 1): 2 / 9,
            (0, 1, 0): 1 / 9,
            (0, 1, 1): 1 / 9,
            (1, 0, 0): 1 / 9,
            (1, 0, 1): 1 / 9,
            (1, 1, 0): 1 / 9,
            (1, 1, 1): 1 / 9,
        }

        remove_file_if_exists(input_filename)

    def test_add_counts(self, counts):
        # Given
        measurements = Measurements()
        measurements_counts = counts

        # When
        measurements.add_counts(measurements_counts)

        # Then
        assert measurements.bitstrings == [
            (0, 0, 0),
            (0, 0, 1),
            (0, 0, 1),
            (0, 1, 0),
            (0, 1, 1),
            (1, 0, 0),
            (1, 0, 1),
            (1, 1, 0),
            (1, 1, 1),
        ]
        assert measurements.get_counts() == counts

    def test_add_measurements(self):
        # Given
        measurements = Measurements()
        bitstrings = [
            (0, 0, 0),
            (0, 0, 1),
            (0, 0, 1),
            (0, 1, 0),
            (0, 1, 1),
            (1, 0, 0),
            (1, 0, 1),
            (1, 1, 0),
            (1, 1, 1),
        ]

        # When
        measurements.bitstrings = bitstrings

        # Then
        assert measurements.bitstrings == [
            (0, 0, 0),
            (0, 0, 1),
            (0, 0, 1),
            (0, 1, 0),
            (0, 1, 1),
            (1, 0, 0),
            (1, 0, 1),
            (1, 1, 0),
            (1, 1, 1),
        ]

        # When
        measurements.bitstrings += bitstrings

        # Then
        assert measurements.bitstrings == [
            (0, 0, 0),
            (0, 0, 1),
            (0, 0, 1),
            (0, 1, 0),
            (0, 1, 1),
            (1, 0, 0),
            (1, 0, 1),
            (1, 1, 0),
            (1, 1, 1),
            (0, 0, 0),
            (0, 0, 1),
            (0, 0, 1),
            (0, 1, 0),
            (0, 1, 1),
            (1, 0, 0),
            (1, 0, 1),
            (1, 1, 0),
            (1, 1, 1),
        ]

    def test_get_expectation_values_from_measurements(self):
        # Given
        measurements = Measurements(
            [(0, 1, 0), (0, 1, 0), (0, 0, 0), (1, 0, 0), (1, 1, 1)]
        )
        ising_operator = (
            PauliTerm("I0", 10) + PauliTerm("1*Z0*Z1") - PauliTerm("15*Z1*Z2")
        )
        target_expectation_values = np.array([10, -0.2, -3])
        target_correlations = np.array([[100, -2, -30], [-2, 1, -9], [-30, -9, 225]])
        denominator = len(measurements.bitstrings)
        covariance_11 = (
            target_correlations[1, 1] - target_expectation_values[1] ** 2
        ) / denominator
        covariance_12 = (
            target_correlations[1, 2]
            - target_expectation_values[1] * target_expectation_values[2]
        ) / denominator
        covariance_22 = (
            target_correlations[2, 2] - target_expectation_values[2] ** 2
        ) / denominator

        target_covariances = np.array(
            [
                [0, 0, 0],
                [0, covariance_11, covariance_12],
                [0, covariance_12, covariance_22],
            ]
        )

        # When
        expectation_values = measurements.get_expectation_values(ising_operator, False)
        # Then
        np.testing.assert_allclose(expectation_values.values, target_expectation_values)
        assert len(expectation_values.correlations) == 1
        np.testing.assert_allclose(
            expectation_values.correlations[0], target_correlations
        )
        assert len(expectation_values.estimator_covariances) == 1
        np.testing.assert_allclose(
            expectation_values.estimator_covariances[0], target_covariances
        )

    def test_get_expectation_values_from_measurements_with_bessel_correction(self):
        # Given
        measurements = Measurements(
            [(0, 1, 0), (0, 1, 0), (0, 0, 0), (1, 0, 0), (1, 1, 1)]
        )
        ising_operator = (
            PauliTerm("I0", 10) + PauliTerm("1*Z0*Z1") - PauliTerm("15*Z1*Z2")
        )
        target_expectation_values = np.array([10, -0.2, -3])
        target_correlations = np.array([[100, -2, -30], [-2, 1, -9], [-30, -9, 225]])
        denominator = len(measurements.bitstrings) - 1
        covariance_11 = (
            target_correlations[1, 1] - target_expectation_values[1] ** 2
        ) / denominator
        covariance_12 = (
            target_correlations[1, 2]
            - target_expectation_values[1] * target_expectation_values[2]
        ) / denominator
        covariance_22 = (
            target_correlations[2, 2] - target_expectation_values[2] ** 2
        ) / denominator

        target_covariances = np.array(
            [
                [0, 0, 0],
                [0, covariance_11, covariance_12],
                [0, covariance_12, covariance_22],
            ]
        )

        # When
        expectation_values = measurements.get_expectation_values(ising_operator, True)
        # Then
        np.testing.assert_allclose(expectation_values.values, target_expectation_values)
        assert len(expectation_values.correlations) == 1
        np.testing.assert_allclose(
            expectation_values.correlations[0], target_correlations
        )
        assert len(expectation_values.estimator_covariances) == 1
        np.testing.assert_allclose(
            expectation_values.estimator_covariances[0], target_covariances
        )

    @pytest.mark.parametrize(
        "bitstring_distribution, number_of_samples",
        [
            (MeasurementOutcomeDistribution({"00": 0.5, "11": 0.5}), 1),
            (MeasurementOutcomeDistribution({"00": 0.5, "11": 0.5}), 10),
            (MeasurementOutcomeDistribution({"00": 0.5, "11": 0.5}), 51),
            (MeasurementOutcomeDistribution({"00": 0.5, "11": 0.5}), 137),
            (MeasurementOutcomeDistribution({"00": 0.5, "11": 0.5}), 5000),
            (MeasurementOutcomeDistribution({"0000": 0.137, "0001": 0.863}), 100),
            (
                MeasurementOutcomeDistribution(
                    {"00": 0.1234, "01": 0.5467, "10": 0.0023, "11": 0.3276}
                ),
                100,
            ),
            (
                MeasurementOutcomeDistribution(
                    {
                        "0000": 0.06835580857498666,
                        "1000": 0.060975627112613416,
                        "0100": 0.05976605586194627,
                        "1100": 0.07138587439957303,
                        "0010": 0.06474168297455969,
                        "1010": 0.0825036470378936,
                        "0110": 0.09861252446183953,
                        "1110": 0.0503013698630137,
                        "0001": 0.04496424123821384,
                        "1001": 0.07317221135029355,
                        "0101": 0.08171161714997331,
                        "1101": 0.03753940579967977,
                        "0011": 0.05157676570005337,
                        "1011": 0.05,
                        "0111": 0.04964419142501335,
                        "1111": 0.05474897705034692,
                    }
                ),
                5621,
            ),
        ],
    )
    def test_get_measurements_representing_distribution_returns_right_number_of_samples(
        self, bitstring_distribution, number_of_samples
    ):
        measurements = Measurements.get_measurements_representing_distribution(
            bitstring_distribution, number_of_samples
        )
        assert len(measurements.bitstrings) == number_of_samples

    @pytest.mark.parametrize(
        "bitstring_distribution, number_of_samples, expected_counts",
        [
            (
                MeasurementOutcomeDistribution(
                    {"01": 0.3333333, "11": (1 - 0.3333333)}
                ),
                3,
                {"01": 1, "11": 2},
            ),
            (
                MeasurementOutcomeDistribution(
                    {"01": 0.9999999, "11": (1 - 0.9999999)}
                ),
                1,
                {"01": 1},
            ),
        ],
    )
    def test_get_measurements_samples_correctly_leftover_bitstrings(
        self, bitstring_distribution, number_of_samples, expected_counts
    ):
        random.seed(RNDSEED)
        measurements = Measurements.get_measurements_representing_distribution(
            bitstring_distribution, number_of_samples
        )
        assert measurements.get_counts() == expected_counts

    def test_get_measurements_samples_correctly_leftover_bitstrings_for_equal_probas(
        self,
    ):
        random.seed(RNDSEED)
        bitstring_distribution = MeasurementOutcomeDistribution({"00": 0.5, "11": 0.5})
        number_of_samples = 51
        max_number_of_trials = 10
        got_different_measurements = False
        previous_measurements = Measurements.get_measurements_representing_distribution(
            bitstring_distribution, number_of_samples
        )

        while not got_different_measurements:
            measurements = Measurements.get_measurements_representing_distribution(
                bitstring_distribution, number_of_samples
            )

            assert measurements.get_counts() == {
                "00": 25,
                "11": 26,
            } or measurements.get_counts() == {"00": 26, "11": 25}

            if measurements.get_counts() != previous_measurements.get_counts():
                got_different_measurements = True

            max_number_of_trials -= 1
            if max_number_of_trials == 0:
                break
        assert got_different_measurements

    @pytest.mark.parametrize(
        "bitstring_distribution",
        [
            MeasurementOutcomeDistribution({"00": 0.5, "11": 0.5}),
            MeasurementOutcomeDistribution({"000": 0.5, "101": 0.5}),
            MeasurementOutcomeDistribution({"0000": 0.137, "0001": 0.863}),
            MeasurementOutcomeDistribution(
                {"00": 0.1234, "01": 0.5467, "10": 0.0023, "11": 0.3276}
            ),
        ],
    )
    def test_get_measurements_representing_distribution_gives_exactly_right_counts(
        self, bitstring_distribution
    ):
        number_of_samples = 10000
        measurements = Measurements.get_measurements_representing_distribution(
            bitstring_distribution, number_of_samples
        )

        counts = measurements.get_counts()
        for bitstring, probability in bitstring_distribution.distribution_dict.items():
            assert (
                probability * number_of_samples
                == counts[convert_tuples_to_bitstrings([bitstring])[0]]
            )

    @pytest.mark.parametrize(
        "bitstring_distribution",
        [
            MeasurementOutcomeDistribution(
                {
                    "0011": 0.0049,
                    "1100": 0.0049,
                    "1111": 0.008,
                    "0000": 0.008,
                    "0001": 0.008,
                    "0010": 0.008,
                    "0100": 0.008,
                    "1000": 0.008,
                    "1001": 0.008,
                    "1010": 0.008,
                    "1110": 0.008,
                    "1101": 0.008,
                    "0110": 0.9102,
                }
            )
        ],
    )
    def test_get_measurements_representing_distribution_doesnt_raise(
        self, bitstring_distribution
    ):
        number_of_samples = 100
        max_number_of_trials = 100
        for _ in range(max_number_of_trials):
            _ = Measurements.get_measurements_representing_distribution(
                bitstring_distribution, number_of_samples
            )

    @pytest.mark.parametrize(
        "samples, bitstring_samples, leftover_distribution",
        [
            (
                Counter({"0011": 3, "1100": 1, "0101": 2}),
                [
                    (0, 0, 1, 1),
                    (1, 1, 0, 0),
                    (0, 1, 0, 1),
                    (0, 1, 0, 1),
                    (0, 1, 0, 1),
                    (1, 1, 0, 0),
                    (0, 0, 1, 1),
                ],
                MeasurementOutcomeDistribution({"0011": 0.3, "1100": 0.3, "0101": 0.3}),
            ),
            (
                Counter({"0011": 3, "1100": 1, "0101": 2, "0001": 1}),
                [
                    (0, 0, 1, 1),
                    (1, 1, 0, 0),
                    (0, 1, 0, 1),
                    (0, 1, 0, 1),
                    (0, 1, 0, 1),
                    (1, 1, 0, 0),
                    (0, 0, 1, 1),
                ],
                MeasurementOutcomeDistribution(
                    {"0011": 0.3, "1100": 0.001, "0101": 0.001, "0001": 0.698}
                ),
            ),
        ],
    )
    def test_check_sample_elimination(
        self, samples, bitstring_samples, leftover_distribution
    ):
        correct_samples = _check_sample_elimination(
            samples, bitstring_samples, leftover_distribution
        )

        bitstring_counts = Counter(bitstring_samples)
        for sample in correct_samples:
            bitstring = tuple([int(measurement_value) for measurement_value in sample])
            assert correct_samples[sample] <= bitstring_counts[bitstring]



################################################################################
# © Copyright 2021-2022 Zapata Computing Inc.
################################################################################
"""Definitions of keys for various symbol orderings."""
import re


def _convert_string_to_int_if_possible(text):
    return int(text) if text.isdigit() else text


def natural_key(symbol):
    """Convert symbol to a natural-ordering key.

    The natural ordering of symbols works as follows:
    1. Split a symbol on each group of digits.
    2. For each group in this split, convert all-digit groups to integers.

    Comparing such groups using lexicographical ordering gives natural ordering
    of symbols.

    In contrast to usual ordering obtained by casting symbol to string, natural
    ordering treats variable indices as integers. Thus, in this ordering,
    symbols beta_10, theta_2, beta_2, theta_1 would be ordered as follows:
    beta_2 < beta_10 < theta_1 < theta_2.

    The original idea behind this implementation was found byMSRudolph here:
    http://nedbatchelder.com/blog/200712/human_sorting.html
    """
    return [
        _convert_string_to_int_if_possible(group)
        for group in re.split(r"(\d+)", symbol.name)
    ]


def natural_key_revlex(symbol):
    """Convert symbol to lexically reversed natural-ordering key.

    This returns reversed key produced by `natural_key`. The main usage is for list
    of symbols with names of the form <symbol_name>_<number> where <symbol_name>
    can take one of several predefined values, and orders should consider <number>
    before <symbol_name>.

    For instance, given symbols beta_1, beta_2, theta_1, theta_2, sorting them
    using natural_key_revlex will give beta_1 < theta_1 < beta_2 < theta_2.
    """
    return list(reversed(natural_key(symbol)))


def natural_key_fixed_names_order(names_order):
    """Convert symbol to natural key but with custom ordering of names.

    Consider a QAOA ansatz in which parameters are naturally ordered as:
    gamma_0 < beta_0 < gamma_1 < beta_1 < ...

    The above is an example of natural_key_fixed_names_order in which name 'gamma'
    precedes name 'beta'.

    Note that unlike natural_key and natural_key_revlex, this function returns
    a key, i.e. it is a key factory.
    """
    symbol_weights = {name: i for i, name in enumerate(names_order)}

    def _key(symbol):
        name, index = symbol.name.split("_")
        return int(index), symbol_weights[name]

    return _key



################################################################################
# © Copyright 2021-2022 Zapata Computing Inc.
################################################################################
from typing import Collection, Optional, Union, cast
from warnings import warn

import numpy as np

from ._builtin_gates import GatePrototype, I
from ._circuit import Circuit
from ._gates import Gate


def create_layer_of_gates(
    number_of_qubits: int,
    gate_factory: Union[GatePrototype, Gate],
    parameters: Optional[np.ndarray] = None,
) -> Circuit:
    """Creates a circuit consisting of a layer of single-qubit gates acting on all
    qubits.

    Args:
        number_of_qubits: number of qubits in the circuit
        gate_factory: the single qubit gate to be applied to each qubit
        parameters: parameters of the single-qubit gates

    Returns:
        circuit: Created circuit.
    """
    circuit = Circuit()

    return apply_gate_to_qubits(
        circuit, range(number_of_qubits), gate_factory, parameters
    )


def apply_gate_to_qubits(
    circuit: Circuit,
    qubit_indices: Collection[int],
    gate_factory: Union[GatePrototype, Gate],
    parameters: Optional[np.ndarray] = None,
) -> Circuit:
    """Apply the passed gate to the specified qubits

    Args:
        circuit: circuit to add the gate to
        qubit_indices: list of qubits that to apply the gate on,
            duplicates will be ignored
        gate_factory: the single qubit gate to be applied to each qubit
        parameters: parameters of the single-qubit gate, this is a list of size
            dependent on the gate used, e.g. 1 for RY gates, 3 for U3 gates
    Returns:
        circuit with the gate added to the specified qubits
    """
    unique_qubit_idx = set(qubit_indices)

    if len(unique_qubit_idx) != len(qubit_indices):
        warn("Duplicate qubits were passed! Duplicates will be ignored.")

    if parameters is not None:
        assert len(parameters) == len(unique_qubit_idx)
        gate_factory = cast(GatePrototype, gate_factory)
        for qubit, parameter in zip(unique_qubit_idx, parameters):
            circuit += gate_factory(*parameter)(qubit)
    else:
        gate_factory = cast(Gate, gate_factory)
        for qubit in unique_qubit_idx:
            circuit += gate_factory(qubit)

    return circuit


def add_ancilla_register(circuit: Circuit, n_ancilla_qubits: int) -> Circuit:
    """Add a register of ancilla qubits (qubit + identity gate) to an existing circuit.

    Args:
        circuit: circuit to be extended
        n_ancilla_qubits: number of ancilla qubits to add
    Returns:
        extended circuit
    """
    extended_circuit = circuit
    for ancilla_qubit_i in range(n_ancilla_qubits):
        qubit_index = circuit.n_qubits + ancilla_qubit_i
        extended_circuit += I(qubit_index)
    return extended_circuit



################################################################################
# © Copyright 2022 Zapata Computing Inc.
################################################################################
import math  # This is needed for tests to work because of monkey patching

from ._measurement_outcome_distribution import (
    MeasurementOutcomeDistribution,
    change_tuple_dict_keys_to_comma_separated_integers,
    create_bitstring_distribution_from_probability_distribution,
    evaluate_distribution_distance,
    is_measurement_outcome_distribution,
    is_normalized,
    load_measurement_outcome_distribution,
    load_measurement_outcome_distributions,
    normalize_measurement_outcome_distribution,
    preprocess_distibution_dict,
    save_measurement_outcome_distribution,
    save_measurement_outcome_distributions,
)

# Distance Measures
from .clipped_negative_log_likelihood import compute_clipped_negative_log_likelihood
from .jensen_shannon_divergence import compute_jensen_shannon_divergence
from .mmd import compute_mmd, compute_multi_rbf_kernel, compute_rbf_kernel



################################################################################
# © Copyright 2021-2022 Zapata Computing Inc.
################################################################################
"""Module containing utilities for handling unitary matrices."""
from functools import reduce

import numpy as np
import sympy


def _permute(vector, permutation):
    """Permute vector according to given ordering of indices.

    Args:
        vector: 1-D array-like object
        permutation: any permutation of indices 0...len(vector)
    Returns:
        permuted vector
    Examples:
        _permute([4, 5, 6], [0, 2, 1]) == [4, 6, 5]
    """
    return [vector[i] for i in permutation]


def _basis_bitstring(i, num_qubits):
    """Create vector corresponding to i-th basis vector of num_qubits system."""
    return [int(char) for char in bin(i)[2:].zfill(num_qubits)]


def _bitstring_to_sympy_dense_vector(state):
    """Construct sympy matrix representation of a state given by a bitstring."""
    basis = [sympy.Matrix([1, 0]), sympy.Matrix([0, 1])]
    return sympy.kronecker_product(*[basis[bit] for bit in state])


def _bitstring_to_numpy_dense_vector(state):
    """Construct numpy array representation of a state given by a bitstring."""
    basis = [np.array([1, 0]), np.array([0, 1])]
    return reduce(np.kron, (basis[bit] for bit in state))


def _permutation_matrix(target_indices_order, zeros, bitstring_to_dense_vector):
    """Construct a permutation matrix for N qubit system.

    Args:
        target_indices_order: the desired order of systems (qubits) after permutation.
           This should contain all entries from 0 to N-1, where N is the number of
           qubits.
        zeros: function for constructing all-zero matrix that will be filled with
            appropriate entries.
        bitstring_to_dense_vector: function constructing dense vector from bitstring.
    Returns:
        Matrix permuting qubits according to given ordering.
    Notes:
        Matrix acts on whole N-qubit system, regardless of how many qubits actually
        change their position. The `zeros` and `bitstring_to_dense_vector` functions are
        used for constructing variants of this function depending on whether sympy or
        numpy should be used (see below how this dispatching is done).
    """
    num_qubits = len(target_indices_order)
    if sorted(target_indices_order) != list(range(num_qubits)):
        raise ValueError("Not all qubits given in permutation.")
    perm_matrix = zeros((2**num_qubits, 2**num_qubits))
    for i in range(2**num_qubits):
        input_state = _basis_bitstring(i, num_qubits)
        output_state = _permute(input_state, target_indices_order)
        perm_matrix[:, i] = bitstring_to_dense_vector(output_state)
    return perm_matrix


def _permutation_making_qubits_adjacent(qubit_indices, num_qubits):
    """Given an iterable of qubit indices construct a permutation such that they are
    next to each other."""
    return list(qubit_indices) + [
        i for i in range(num_qubits) if i not in qubit_indices
    ]


def _lift_matrix(
    matrix,
    qubit_indices,
    num_qubits,
    zeros,
    eye,
    kronecker_product,
    bitstring_to_dense_vector,
):
    """Lift a matrix acting on subsystem of N-qubit system to one acting on the whole
    system.

    Args:
        matrix: matrix acting on k `qubits`
        qubit_indices: indices of qubits that matrix acts on
        zeros: function constructing all-zero matrix. It is assumed that it can be
            called like `zeros((m, n))`.
        eye: function constructing identity matrix
        kronecker_product: function computing kronecker product of matrices.
            It is assumed that it can be applied like `kronecker_product(x, y)`.
        bitstring_to_dense_vector: function converting bitstring to dense vector.
    Returns:
        Matrix that acts like `matrix` on qubits with indices `qubit_indices` and as
            identity on other qubits.
    """
    smallest, largest = min(qubit_indices), max(qubit_indices)
    # No need to consider all the qubits, just those between smallest and largest one
    shifted_qubits = [index - smallest for index in qubit_indices]
    inner_permutation = _permutation_making_qubits_adjacent(
        shifted_qubits, largest - smallest + 1
    )
    # perm_matrix permutes qubits in range smallest-largest so that the active
    # ones come first.
    perm_matrix = _permutation_matrix(
        inner_permutation, zeros, bitstring_to_dense_vector
    )

    # inner_gate_matrix acts on the whole range smallest-largest, by
    # transforming first qubits in the same way as matrix, and leaving
    # others unchanged
    inner_gate_matrix = kronecker_product(
        matrix, eye(2 ** (largest - smallest - len(qubit_indices) + 1))
    )
    # target operation acting on smallest-largest range is now composed of
    # 1. basis change (via permutation)
    # 2. gate action
    # 3. reversal of basis change
    inner_matrix = perm_matrix.transpose() @ inner_gate_matrix @ perm_matrix
    # finally, to make a matrix acting on whole range of qubits, we
    # add identities acting on qubits with indices outside of smallest-largest range.
    return reduce(
        kronecker_product,
        [eye(2**smallest), inner_matrix, eye(2 ** (num_qubits - largest - 1))],
    )


def _lift_matrix_numpy(matrix, qubits, num_qubits):
    """A version of _lift_matrix working on numpy arrays.

    Notice that the input matrix is typically sympy's matrix, so we have to first
    convert it.
    """
    matrix = np.array(matrix, dtype=complex)
    return _lift_matrix(
        matrix,
        qubits,
        num_qubits,
        np.zeros,
        np.eye,
        np.kron,
        _bitstring_to_numpy_dense_vector,
    )


def _lift_matrix_sympy(matrix, qubits, num_qubits):
    """A version of _lift_matrix working on Sympy matrices."""
    return _lift_matrix(
        matrix,
        qubits,
        num_qubits,
        # sympy has different signature of `zeros` than numpy, hence the below adapter.
        lambda size: sympy.zeros(*size),
        sympy.eye,
        sympy.kronecker_product,
        _bitstring_to_sympy_dense_vector,
    )



################################################################################
# © Copyright 2022 Zapata Computing Inc.
################################################################################
import json
from collections import Counter
from typing import Any, Dict, Iterable, List, Optional, Sequence, Tuple, Union

import numpy as np

from ..operators import PauliRepresentation
from ..typing import AnyPath, LoadSource
from ..utils import convert_array_to_dict, convert_dict_to_array, ensure_open


class Parities:
    """A class representing counts of parities for Pauli terms.

    Args:
        values (np.array): Number of observations of parities. See Attributes.
        correlations (list): Number of observations of pairwise products of terms.
            See Attributes.

    Attributes:
        values (np.array): an array of dimension N x 2 indicating how many times
            each Pauli term was observed with even and odd parity, where N is the
            number of Pauli terms. Here values[i][0] and values[i][1] correspond
            to the number of samples with even and odd parities for term P_i,
            respectively.
        correlations (list): a list of 3-dimensional numpy arrays indicating how
            many times each product of Pauli terms was observed with even and odd
            parity. Here correlations[i][j][k][0] and correlations[i][j][k][1]
            correspond to the number of samples with even and odd parities term P_j P_k
            in frame i, respectively.
    """

    def __init__(
        self, values: np.ndarray, correlations: Optional[List[np.ndarray]] = None
    ):
        self.values = values
        self.correlations = correlations

    def to_dict(self) -> dict:
        data: Dict[str, Any] = {"values": convert_array_to_dict(self.values)}
        if self.correlations:
            data["correlations"] = [
                convert_array_to_dict(arr) for arr in self.correlations
            ]
        return data

    @classmethod
    def from_dict(cls, data: dict):
        values = convert_dict_to_array(data["values"])
        if data.get("correlations"):
            correlations: Optional[List] = [
                convert_dict_to_array(arr) for arr in data["correlations"]
            ]
        else:
            correlations = None
        return cls(values, correlations)


def save_parities(parities: Parities, filename: AnyPath) -> None:
    """Save parities to a file.

    Args:
        parities (orquestra.quantum.measurement.Parities): the parities
        file (str or file-like object): the name of the file, or a file-like object
    """
    data = parities.to_dict()

    with open(filename, "w") as f:
        f.write(json.dumps(data, indent=2))


def load_parities(file: LoadSource) -> Parities:
    """Load parities from a file.

    Args:
        file (str or file-like object): the name of the file, or a file-like object.

    Returns:
        orquestra.quantum.measurement.Parities: the parities
    """

    with ensure_open(file) as f:
        data = json.load(f)

    return Parities.from_dict(data)


def check_parity(
    bitstring: Union[str, Sequence[int]], marked_qubits: Iterable[int]
) -> bool:
    """Determine if the marked qubits have even parity for the given bitstring.

    Args:
        bitstring: The bitstring, either as a tuple or little endian string.
        marked_qubits: The qubits whose parity is to be determined.

    Returns:
        True if an even number of the marked qubits are in the 1 state, False
            otherwise.
    """
    result = True
    for qubit_index in marked_qubits:
        if bitstring[qubit_index] == "1" or bitstring[qubit_index] == 1:
            result = not result
    return result


def check_parity_of_vector(
    bitstrings_vector: np.ndarray, marked_qubits: Iterable[int]
) -> np.ndarray:
    """Determine if the marked qubits have even parity for each bitstring in the given
        vector.
        NOTE: This performs the same functionality as `check_parity` but is much
        faster as it uses vectorization to find the parity of multiple bitstrings
        at once.

    Args:
        bitstring: A 2d array of bitstrings whose size is number of bistrings * number
            of qubits
        marked_qubits: The qubits whose parity is to be determined.

    Returns:
        A 1d array with size equal to number of bitstrings. Each entry is 1 if an even
        number of the marked qubits of the corresponding bitstring are in the 1 state
        and 0 if otherwise.
    """
    if not marked_qubits:
        return np.ones(bitstrings_vector.shape[0])

    # Check if an even number of the marked qubits of each bitstring are in the 1 state
    bitstring_subset = bitstrings_vector[:, np.fromiter(marked_qubits, dtype=int)]
    return (bitstring_subset.sum(axis=1) + 1) % 2


def get_parities_from_measurements(
    measurements: List[Tuple[int]], ising_operator: PauliRepresentation
) -> Parities:
    """Get expectation values from bitstrings.

    Args:
        measurements (list): the measured bitstrings
        ising_operator: the operator

    Returns:
        orquestra.quantum.measurement.Parities: parities of each term in the operator
    """

    # check input format
    if not ising_operator.is_ising:
        raise TypeError("Input operator is not an ising operator")

    # Count number of occurrences of bitstrings
    bitstring_frequencies = Counter(measurements)
    bitstrings_vector = np.array([*bitstring_frequencies.keys()])
    bitstring_counts: np.ndarray = np.fromiter(
        bitstring_frequencies.values(), dtype=int
    )

    # Count parity occurrences
    values = []
    for _, term in enumerate(ising_operator.terms):
        parity = check_parity_of_vector(bitstrings_vector, term.qubits)

        true_parity_count: int = (parity * bitstring_counts).sum()
        false_parity_count: int = ((1 - parity) * bitstring_counts).sum()
        values.append([true_parity_count, false_parity_count])

    # Count parity occurrences for pairwise products of operators
    correlations = [np.zeros((len(ising_operator.terms), len(ising_operator.terms), 2))]
    for term1_index, term1 in enumerate(ising_operator.terms):
        for term2_index, term2 in enumerate(ising_operator.terms):

            parity1 = check_parity_of_vector(bitstrings_vector, term1.qubits)
            parity2 = check_parity_of_vector(bitstrings_vector, term2.qubits)

            # 0 if parities are equal, 1 otherwise
            equal_parities = np.abs(parity1 - parity2)

            # Counts of bitstrings where parity is equal
            correlations[0][term1_index, term2_index][0] += (
                (1 - equal_parities) * bitstring_counts
            ).sum()

            # Counts of bitstrings where parity is not equal
            correlations[0][term1_index, term2_index][1] += (
                (equal_parities) * bitstring_counts
            ).sum()

    return Parities(np.array(values), correlations)



################################################################################
# © Copyright 2020-2022 Zapata Computing Inc.
################################################################################

import numpy as np

from ..wavefunction import Wavefunction


def create_random_wavefunction(n_qubits, seed=None):
    """Create a random wavefunction for testing purposes."""
    if seed:
        np.random.seed(seed)

    random_vector = [
        complex(a, b)
        for a, b in zip(np.random.rand(2**n_qubits), np.random.rand(2**n_qubits))
    ]
    normalization_factor = np.sqrt(np.sum(np.abs(random_vector) ** 2))
    random_vector /= normalization_factor

    return Wavefunction(random_vector)



################################################################################
# © Copyright 2021-2022 Zapata Computing Inc.
################################################################################
"""Utilities related to translation of symbolic expressions."""
from functools import singledispatch
from numbers import Number
from typing import Iterable, Tuple, Union

from .expressions import Expression, ExpressionDialect, FunctionCall, Symbol


@singledispatch
def translate_expression(
    expression: Union[Expression, Tuple[Expression, ...]], dialect: ExpressionDialect
):
    pass


@translate_expression.register
def translate_number(number: Number, dialect: ExpressionDialect):
    return dialect.number_factory(number)


@translate_expression.register
def translate_symbol(symbol: Symbol, dialect: ExpressionDialect):
    return dialect.symbol_factory(symbol)


@translate_expression.register
def translate_function_call(function_call: FunctionCall, dialect: ExpressionDialect):
    if function_call.name not in dialect.known_functions:
        raise ValueError(f"Function {function_call.name} is unknown in this dialect.")

    return dialect.known_functions[function_call.name](
        *translate_tuple(function_call.args, dialect)
    )


def translate_tuple(expression_tuple: Iterable[Expression], dialect: ExpressionDialect):
    return tuple(translate_expression(element, dialect) for element in expression_tuple)



################################################################################
# © Copyright 2021-2022 Zapata Computing Inc.
################################################################################
from typing import Dict, List, Optional, Tuple, cast

import numpy as np
import sympy

from ..api.circuit_runner import CircuitRunner
from ..api.estimation import EstimationTask
from ..api.wavefunction_simulator import WavefunctionSimulator
from ..measurements import ExpectationValues, expectation_values_to_real


def evaluate_estimation_circuits(
    estimation_tasks: List[EstimationTask],
    symbols_maps: List[Dict[sympy.Symbol, float]],
) -> List[EstimationTask]:
    """Evaluates circuits given in all estimation tasks using the given symbols_maps.

    If one symbols map is given, it is used to evaluate all circuits. Otherwise, the
    symbols map at index i will be used for the estimation task at index i.

    Args:
        estimation_tasks: the estimation tasks which contain the circuits to be
            evaluated
        symbols_maps: a list of dictionaries (or singular dictionary) that map the
            symbolic symbols used in the parametrized circuits to the associated values
    """
    return [
        EstimationTask(
            operator=estimation_task.operator,
            circuit=estimation_task.circuit.bind(symbols_map),
            number_of_shots=estimation_task.number_of_shots,
        )
        for estimation_task, symbols_map in zip(estimation_tasks, symbols_maps)
    ]


def split_estimation_tasks_to_measure(
    estimation_tasks: List[EstimationTask],
) -> Tuple[List[EstimationTask], List[EstimationTask], List[int], List[int]]:
    """This function splits a given list of EstimationTask into two: one that
    contains EstimationTasks that should be measured, and one that contains
    EstimationTasks with constants or with 0 shots.

    Args:
        estimation_tasks: The list of estimation tasks for which
                         Expectation Values are wanted.

    Returns:
        estimation_tasks_to_measure: A new list of estimation tasks that only
            contains the ones that should actually be submitted to the runner
        estimation_tasks_not_to_measure: A new list of estimation tasks that
            contains the EstimationTasks with only constant terms or with
            0 shot
        indices_to_measure: A list containing the indices of the EstimationTasks we will
            actually measure, i.e. the ith estimation_tasks_to_measure expectation
            value will go into the indices_to_measure[i] position.
        indices_not_to_measure: A list containing the indices of the EstimationTasks for
            constant terms or with 0 shot.
    """

    estimation_tasks_to_measure = []
    estimation_tasks_not_to_measure = []
    indices_to_measure = []
    indices_not_to_measure = []
    for i, task in enumerate(estimation_tasks):
        if task.operator.is_constant or task.number_of_shots == 0:
            indices_not_to_measure.append(i)
            estimation_tasks_not_to_measure.append(task)
        else:
            indices_to_measure.append(i)
            estimation_tasks_to_measure.append(task)

    return (
        estimation_tasks_to_measure,
        estimation_tasks_not_to_measure,
        indices_to_measure,
        indices_not_to_measure,
    )


def evaluate_non_measured_estimation_tasks(
    estimation_tasks: List[EstimationTask],
) -> List[ExpectationValues]:
    """This function evaluates a list of EstimationTask that are not
    measured, and either contain only a constant term or require 0 shot.
    Non-constant EstimationTask with 0 shot return 0.0 as their
    ExpectationValue, with a precision of 0.0 as well.

    Args:
        estimation_tasks: The list of estimation tasks for which
            Expectation Values are wanted.

    Returns:
        expectation_values: the expectation values over non-measured terms,
            with their correlations and estimator_covariances.
    """

    expectation_values = []
    for task in estimation_tasks:
        coefficient: complex
        if task.operator.is_constant:
            coefficient = task.operator.terms[0].coefficient
        else:
            if task.number_of_shots is not None and task.number_of_shots > 0:
                raise RuntimeError(
                    "An EstimationTask required shots but was classified as "
                    "a non-measured task"
                )
            else:
                coefficient = 0.0

        expectation_values.append(
            ExpectationValues(
                np.asarray([coefficient]),
                correlations=[np.asarray([[0.0]])],
                estimator_covariances=[np.asarray([[0.0]])],
            )
        )

    return expectation_values


def estimate_expectation_values_by_averaging(
    runner: CircuitRunner,
    estimation_tasks: List[EstimationTask],
) -> List[ExpectationValues]:
    """Basic method for estimating expectation values for list of estimation tasks.

    It executes specified circuit and calculates expectation values based on the
    measurements.

    Args:
        runner: runner used for executing circuits
        estimation_tasks: list of estimation tasks
    """

    (
        estimation_tasks_to_measure,
        estimation_tasks_not_to_measure,
        indices_to_measure,
        indices_not_to_measure,
    ) = split_estimation_tasks_to_measure(estimation_tasks)

    non_measured_expectation_values_list = evaluate_non_measured_estimation_tasks(
        estimation_tasks_not_to_measure
    )

    if not estimation_tasks_to_measure:
        measured_expectation_values_list = []
    else:
        circuits, operators, shots_per_circuit = zip(
            *[
                (e.circuit, e.operator, e.number_of_shots)
                for e in estimation_tasks_to_measure
            ]
        )
        measurements_list = runner.run_batch_and_measure(circuits, shots_per_circuit)

        measured_expectation_values_list = [
            expectation_values_to_real(
                measurements.get_expectation_values(frame_operator)
            )
            for frame_operator, measurements in zip(operators, measurements_list)
        ]

    full_expectation_values: List[Optional[ExpectationValues]] = [
        None
        for _ in range(
            len(estimation_tasks_not_to_measure) + len(estimation_tasks_to_measure)
        )
    ]

    for ex_val, final_index in zip(
        non_measured_expectation_values_list, indices_not_to_measure
    ):
        full_expectation_values[final_index] = ex_val
    for ex_val, final_index in zip(
        measured_expectation_values_list, indices_to_measure
    ):
        full_expectation_values[final_index] = ex_val

    return cast(List[ExpectationValues], full_expectation_values)


def calculate_exact_expectation_values(
    runner: WavefunctionSimulator,
    estimation_tasks: List[EstimationTask],
) -> List[ExpectationValues]:
    """Calculates exact expectation values using built-in method of a provided runner.

    Args:
        runner: runner used for executing circuits
        estimation_tasks: list of estimation tasks
    """
    expectation_values_list = [
        runner.get_exact_expectation_values(
            estimation_task.circuit, estimation_task.operator
        )
        for estimation_task in estimation_tasks
    ]
    return [ExpectationValues(np.asarray([val])) for val in expectation_values_list]



################################################################################
# © Copyright 2022 Zapata Computing Inc.
################################################################################
from functools import partial
from itertools import chain
from typing import List, Tuple

import numpy as np
import pytest

from orquestra.quantum.operators import PauliRepresentation, PauliSum

from ..circuits import CNOT, RX, RY, RZ, Circuit, H, builtin_gate_by_name
from ..estimation import estimate_expectation_values_by_averaging
from ..operators import PauliTerm
from ..testing.test_cases_for_backend_tests import (
    one_qubit_non_parametric_gates_exp_vals_test_set,
    one_qubit_parametric_gates_exp_vals_test_set,
    two_qubit_non_parametric_gates_exp_vals_test_set,
    two_qubit_parametric_gates_exp_vals_test_set,
)
from . import EstimationTask
from .circuit_runner import CircuitRunner

_EXAMPLE_CIRCUITS = (
    Circuit([RY(np.pi / 5)(0)]),
    Circuit([H(0), CNOT(0, 1)]),
    Circuit([H(0), CNOT(0, 2), H(1)]),
    Circuit([RX(np.pi / 2)(0), H(2), RZ(np.pi / 3)(1), CNOT(0, 3)]),
)

_EXAMPLE_N_SAMPLES = (5, 10, 15, 20)

_INVALID_N_SAMPLES = (0, -1, -10)


class _ValidateRunAndMeasure:
    """Contracts relating to run_and_measure."""

    @staticmethod
    def returns_number_of_measurements_greater_or_equal_to_n_samples(
        runner: CircuitRunner,
    ):
        circuit = _EXAMPLE_CIRCUITS[0]

        def _count_num_measurements(n_samples):
            return len(runner.run_and_measure(circuit, n_samples=n_samples).bitstrings)

        return all(
            _count_num_measurements(n_samples) >= n_samples
            for n_samples in _EXAMPLE_N_SAMPLES
        )

    @staticmethod
    def returns_bitstrings_with_length_equal_to_number_of_qubits_in_circuit(
        runner: CircuitRunner,
    ):
        return all(
            len(bitstring) == circuit.n_qubits
            for circuit in _EXAMPLE_CIRCUITS
            for bitstring in runner.run_and_measure(circuit, n_samples=10).bitstrings
        )

    @staticmethod
    def raises_value_error_if_n_samples_is_nonpositive(runner: CircuitRunner):
        for n_samples in _INVALID_N_SAMPLES:
            try:
                runner.run_and_measure(_EXAMPLE_CIRCUITS[0], n_samples=n_samples)
                return False
            except ValueError:
                pass
            except Exception:
                return False

        return True


class _ValidateRunBatchAndMeasure:
    @staticmethod
    def returns_measurement_object_for_each_circuit_in_batch(runner: CircuitRunner):
        return len(
            runner.run_batch_and_measure(_EXAMPLE_CIRCUITS, _EXAMPLE_N_SAMPLES)
        ) == len(_EXAMPLE_CIRCUITS)

    class ReturnsNumberOfMeasurementsEqualToNSamples:
        @staticmethod
        def when_n_samples_different_for_each_circuit(runner: CircuitRunner):
            return all(
                len(measurement.bitstrings) >= n_samples
                for measurement, n_samples in zip(
                    runner.run_batch_and_measure(_EXAMPLE_CIRCUITS, _EXAMPLE_N_SAMPLES),
                    _EXAMPLE_N_SAMPLES,
                )
            )

        @staticmethod
        def when_n_samples_same_for_each_circuit(runner: CircuitRunner):
            return all(
                len(measurements.bitstrings) >= 10
                for measurements in runner.run_batch_and_measure(
                    _EXAMPLE_CIRCUITS, n_samples=10
                )
            )

    @staticmethod
    def returns_bitstrings_with_length_equal_to_number_of_qubits_in_circuit(
        runner: CircuitRunner,
    ):
        return all(
            len(bitstring) == circuit.n_qubits
            for measurements, circuit in zip(
                runner.run_batch_and_measure(_EXAMPLE_CIRCUITS, _EXAMPLE_N_SAMPLES),
                _EXAMPLE_CIRCUITS,
            )
            for bitstring in measurements.bitstrings
        )

    @staticmethod
    def raises_value_error_if_n_samples_is_nonpositive(runner: CircuitRunner):
        def _for_all_circuits():
            try:
                runner.run_batch_and_measure(
                    _EXAMPLE_CIRCUITS, n_samples=_INVALID_N_SAMPLES
                )
                return False
            except ValueError:
                pass
            except Exception:
                return False

            return True

        def _for_at_least_one_circuit():
            n_samples = (-1,) + (len(_EXAMPLE_CIRCUITS) - 1) * (10,)
            try:
                runner.run_batch_and_measure(_EXAMPLE_CIRCUITS, n_samples=n_samples)
                return False
            except ValueError:
                pass
            except Exception:
                return False

            return True

        return _for_all_circuits() and _for_at_least_one_circuit()

    @staticmethod
    def raises_value_error_if_len_of_n_samples_does_not_match_len_of_batch(
        runner: CircuitRunner,
    ):
        invalid_n_samples = (
            (len(_EXAMPLE_CIRCUITS) + 1) * (10,),
            (len(_EXAMPLE_CIRCUITS) - 1) * (10,),
        )

        for n_samples in invalid_n_samples:
            try:
                runner.run_batch_and_measure(_EXAMPLE_CIRCUITS, n_samples)
                return False
            except ValueError:
                pass
            except Exception:
                return False

        return True


class _ValidateMeasurementOutcomeDistribution:
    @staticmethod
    def returns_distribution_with_number_of_bits_same_as_circuit_n_qubits(
        runner: CircuitRunner,
    ):
        return all(
            (
                runner.get_measurement_outcome_distribution(
                    circuit, n_samples=10
                ).get_number_of_subsystems()
                == circuit.n_qubits
            )
            for circuit in _EXAMPLE_CIRCUITS
        )

    @staticmethod
    def raises_value_error_if_n_samples_is_nonpositive(runner: CircuitRunner):
        for n_samples in _INVALID_N_SAMPLES:
            try:
                runner.get_measurement_outcome_distribution(
                    _EXAMPLE_CIRCUITS[0], n_samples=n_samples
                )
                return False
            except ValueError:
                pass
            except Exception:
                return False

        return True


def strict_runner_returns_number_of_measurements_equal_to_n_samples(
    runner: CircuitRunner,
):
    def _when_n_samples_is_the_same_for_each_circuit():
        return all(
            len(measurements.bitstrings) == 10
            for measurements in runner.run_batch_and_measure(
                _EXAMPLE_CIRCUITS, n_samples=10
            )
        )

    def _when_n_samples_is_different_for_each_circuit():
        return all(
            len(measurement.bitstrings) == n_samples
            for measurement, n_samples in zip(
                runner.run_batch_and_measure(_EXAMPLE_CIRCUITS, _EXAMPLE_N_SAMPLES),
                _EXAMPLE_N_SAMPLES,
            )
        )

    # We make sure to run each subtest, otherwise short-circuiting of
    # circuits might result in missing some errors
    subtests = [
        _when_n_samples_is_different_for_each_circuit(),
        _when_n_samples_is_the_same_for_each_circuit(),
    ]
    return all(subtests)


# TODO: remove this code when it is no longer needed
# This code is temporarily copied from orquestra-vqa to facilitate estimating
# expectation values of non-ising operators. It should be removed once we
# decide what to do with the gate-compatibility contracts.


def get_context_selection_circuit_for_group(
    qubit_operator: PauliRepresentation,
) -> Tuple[Circuit, PauliSum]:
    """Get the context selection circuit for measuring the expectation value
    of a group of co-measurable Pauli terms.
    Args:
        qubit_operator: operator representing group of co-measurable Pauli term
    """
    context_selection_circuit = Circuit()
    transformed_operator = PauliSum([])
    context: List[Tuple[str, int]] = []

    for term in qubit_operator.terms:
        term_operator = PauliTerm.identity()
        for qubit, operator in term.operations:
            for existing_qubit, existing_operator in context:
                if existing_qubit == qubit and existing_operator != operator:
                    raise ValueError("Terms are not co-measurable")
            if (operator, qubit) not in context:
                context.append((operator, qubit))
            product = term_operator * PauliTerm({qubit: "Z"})
            assert isinstance(product, PauliTerm)
            term_operator = product
        transformed_operator += term_operator * term.coefficient

    for factor in context:
        if factor[0] == "X":
            context_selection_circuit += RY(-np.pi / 2)(factor[1])
        elif factor[0] == "Y":
            context_selection_circuit += RX(np.pi / 2)(factor[1])

    return context_selection_circuit, transformed_operator


def perform_context_selection(
    estimation_tasks: List[EstimationTask],
) -> List[EstimationTask]:
    """Changes the circuits in estimation tasks to involve context selection.
    Args:
        estimation_tasks: list of estimation tasks
    """
    output_estimation_tasks = []
    for estimation_task in estimation_tasks:
        (
            context_selection_circuit,
            frame_operator,
        ) = get_context_selection_circuit_for_group(estimation_task.operator)
        frame_circuit = estimation_task.circuit + context_selection_circuit
        new_estimation_task = EstimationTask(
            frame_operator, frame_circuit, estimation_task.number_of_shots
        )
        output_estimation_tasks.append(new_estimation_task)
    return output_estimation_tasks


# ----- End of code copied from orquestra-vqa


def _verify_expectation_value_by_averaging(
    runner, circuit, operator, target_value, n_samples, exp_val_spread
):
    estimation_tasks = perform_context_selection(
        [EstimationTask(operator, circuit, n_samples)]
    )

    calculated_value = estimate_expectation_values_by_averaging(
        runner, estimation_tasks
    )

    sigma = 1 / np.sqrt(n_samples)

    return calculated_value[0].values[0] == pytest.approx(
        target_value, abs=exp_val_spread * sigma * 3
    )


def _one_qubit_nonparametric_gate_test_cases(gates_to_exclude):
    operators = [
        PauliTerm.identity(),
        PauliTerm("X0"),
        PauliTerm("Y0"),
        PauliTerm("Z0"),
    ]
    return [
        (
            Circuit(
                [
                    builtin_gate_by_name(initial_gate)(0),
                    builtin_gate_by_name(tested_gate)(0),
                ]
            ),
            operator,
            target_value,
        )
        for initial_gate, tested_gate, target_values in one_qubit_non_parametric_gates_exp_vals_test_set  # noqa: E501
        for operator, target_value in zip(operators, target_values)
        if tested_gate not in gates_to_exclude
    ]


def _one_qubit_parametric_gate_test_cases(gates_to_exclude):
    operators = [
        PauliTerm.identity(),
        PauliTerm("X0"),
        PauliTerm("Y0"),
        PauliTerm("Z0"),
    ]
    return [
        (
            Circuit(
                [
                    builtin_gate_by_name(initial_gate)(0),
                    builtin_gate_by_name(tested_gate)(*params)(0),
                ]
            ),
            operator,
            target_value,
        )
        for initial_gate, tested_gate, params, target_values in one_qubit_parametric_gates_exp_vals_test_set  # noqa: E501
        for operator, target_value in zip(operators, target_values)
        if tested_gate not in gates_to_exclude
    ]


def _two_qubit_nonparametric_gate_test_cases(gates_to_exclude):
    return [
        (
            Circuit(
                [
                    builtin_gate_by_name(initial_gates[0])(0),
                    builtin_gate_by_name(initial_gates[1])(1),
                    builtin_gate_by_name(tested_gate)(0, 1),
                ]
            ),
            PauliTerm(operator),
            target_value,
        )
        for initial_gates, tested_gate, operators, target_values in two_qubit_non_parametric_gates_exp_vals_test_set  # noqa: E501
        for operator, target_value in zip(operators, target_values)
        if tested_gate not in gates_to_exclude
    ]


def _two_qubit_parametric_gate_test_cases(gates_to_exclude):
    return [
        (
            Circuit(
                [
                    builtin_gate_by_name(initial_gates[0])(0),
                    builtin_gate_by_name(initial_gates[1])(1),
                    builtin_gate_by_name(tested_gate)(*params)(0, 1),
                ]
            ),
            PauliTerm(operator),
            target_value,
        )
        for initial_gates, tested_gate, operators, params, target_values in two_qubit_parametric_gates_exp_vals_test_set  # noqa: E501
        for operator, target_value in zip(operators, target_values)
        if tested_gate not in gates_to_exclude
    ]


CIRCUIT_RUNNER_CONTRACTS = [
    _ValidateRunAndMeasure.returns_number_of_measurements_greater_or_equal_to_n_samples,  # noqa: E501
    _ValidateRunAndMeasure.returns_bitstrings_with_length_equal_to_number_of_qubits_in_circuit,  # noqa: E501
    _ValidateRunAndMeasure.raises_value_error_if_n_samples_is_nonpositive,  # noqa: E501
    _ValidateRunBatchAndMeasure.ReturnsNumberOfMeasurementsEqualToNSamples.when_n_samples_different_for_each_circuit,  # noqa: E501
    _ValidateRunBatchAndMeasure.ReturnsNumberOfMeasurementsEqualToNSamples.when_n_samples_same_for_each_circuit,  # noqa: E501
    _ValidateRunBatchAndMeasure.returns_bitstrings_with_length_equal_to_number_of_qubits_in_circuit,  # noqa: E501
    _ValidateRunBatchAndMeasure.returns_measurement_object_for_each_circuit_in_batch,  # noqa: E501
    _ValidateRunBatchAndMeasure.raises_value_error_if_n_samples_is_nonpositive,  # noqa: E501
    _ValidateRunBatchAndMeasure.raises_value_error_if_len_of_n_samples_does_not_match_len_of_batch,  # noqa: E501
    _ValidateMeasurementOutcomeDistribution.returns_distribution_with_number_of_bits_same_as_circuit_n_qubits,  # noqa: E501
    _ValidateMeasurementOutcomeDistribution.raises_value_error_if_n_samples_is_nonpositive,  # noqa: E501
]


# This set of contracts is meant to be used for runners that return exactly the
# requested number of samples when running circuits. For context: some circuit
# runners return number of samples larger than the requested one, usually
# as a result of batching.
STRICT_CIRCUIT_RUNNER_CONTRACTS = [
    strict_runner_returns_number_of_measurements_equal_to_n_samples
]


def circuit_runner_gate_compatibility_contracts(
    exp_val_spread=1.0, gates_to_exclude=None
):
    gates_to_exclude = [] if gates_to_exclude is None else gates_to_exclude
    n_samples = 1000

    return [
        partial(
            _verify_expectation_value_by_averaging,
            circuit=circuit,
            operator=operator,
            target_value=target_value,
            n_samples=n_samples,
            exp_val_spread=exp_val_spread,
        )
        for circuit, operator, target_value in chain(
            _one_qubit_nonparametric_gate_test_cases(gates_to_exclude),
            _one_qubit_parametric_gate_test_cases(gates_to_exclude),
            _two_qubit_nonparametric_gate_test_cases(gates_to_exclude),
        )
    ]



################################################################################
# © Copyright 2021-2022 Zapata Computing Inc.
################################################################################
import json
import re
from functools import singledispatch
from typing import Dict, Iterable, List, Mapping, Union

import sympy

from ..typing import DumpTarget, LoadSource
from ..utils import ensure_open
from . import _builtin_gates, _circuit, _gates


def serialize_expr(expr: sympy.Expr):
    return str(expr)


def _make_symbols_map(
    symbol_names: Iterable[str],
) -> Dict[str, Union[sympy.Symbol, Dict[int, sympy.Symbol]]]:
    symbols_map: Dict[str, Union[sympy.Symbol, Dict[int, sympy.Symbol]]] = {}
    for name in symbol_names:
        # Check if the symbol name has brackets, such as "x[4]". Such symbol names will
        # occur in circuits imported from Qiskit when the Qiskit circuit was
        # parameterized using a Qiskit ParameterVector.
        match = re.search(r"^(.*)\[([0-9]+)\]$", name)
        if match:
            symbols_map.setdefault(match.group(1), {})[
                int(match.group(2))
            ] = sympy.Symbol(name)
        else:
            symbols_map[name] = sympy.Symbol(name)

    return symbols_map


def deserialize_expr(expr_str, symbol_names):
    symbols_map = _make_symbols_map(symbol_names)
    return sympy.sympify(expr_str, locals=symbols_map)


def builtin_gate_by_name(name):
    return _builtin_gates.builtin_gate_by_name(name)


def _matrix_to_json(matrix: sympy.Matrix):
    return [
        [serialize_expr(element) for element in matrix.row(row_i)]
        for row_i in range(matrix.shape[0])
    ]


def _matrix_from_json(
    json_rows: List[List[str]], symbols_names: Iterable[str]
) -> sympy.Matrix:
    return sympy.Matrix(
        [
            [deserialize_expr(element, symbols_names) for element in json_row]
            for json_row in json_rows
        ]
    )


def _map_eager(fn, iterable: Iterable):
    return list(map(fn, iterable))


# ---------- serialization ----------


@singledispatch
def to_dict(obj):
    raise NotImplementedError(f"Serialization isn't implemented for {type(obj)}")


@to_dict.register
def _circuit_to_dict(circuit: _circuit.Circuit):
    """
    Returns:
        A mapping with keys:
            - "n_qubits"
            - "symbolic_params"
            - "gates"
    """
    custom_gate_definitions = circuit.collect_custom_gate_definitions()
    return {
        "n_qubits": circuit.n_qubits,
        **(
            {
                "operations": _map_eager(to_dict, circuit.operations),
            }
            if circuit.operations
            else {}
        ),
        **(
            {
                "custom_gate_definitions": _map_eager(to_dict, custom_gate_definitions),
            }
            if custom_gate_definitions
            else {}
        ),
    }


@to_dict.register(list)
def _circuitset_to_dict(circuitset: List[_circuit.Circuit]) -> Mapping:
    """
    Returns:
        A mapping with keys:
            - "circuits" - list of circuits in this circuitset
    """
    return {
        "circuits": _map_eager(_circuit_to_dict, circuitset),
    }


@to_dict.register
def _gate_operation_to_dict(gate_operation: _gates.GateOperation):
    return {
        "type": "gate_operation",
        "gate": to_dict(gate_operation.gate),
        "qubit_indices": list(gate_operation.qubit_indices),
    }


@to_dict.register
def _basic_gate_to_dict(gate: _gates.MatrixFactoryGate):
    return {
        "name": gate.name,
        **({"params": _map_eager(serialize_expr, gate.params)} if gate.params else {}),
        **(
            {"free_symbols": sorted(map(str, gate.free_symbols))}
            if gate.free_symbols
            else {}
        ),
    }


@to_dict.register
def _custom_gate_def_to_dict(gate_def: _gates.CustomGateDefinition):
    return {
        "gate_name": gate_def.gate_name,
        "matrix": _matrix_to_json(gate_def.matrix),
        "params_ordering": _map_eager(serialize_expr, gate_def.params_ordering),
    }


@to_dict.register
def _controlled_gate_to_dict(gate: _gates.ControlledGate):
    return {
        "name": gate.name,
        "wrapped_gate": to_dict(gate.wrapped_gate),
        "num_control_qubits": gate.num_control_qubits,
    }


@to_dict.register
def _dagger_gate_to_dict(gate: _gates.Dagger):
    return {
        "name": gate.name,
        "wrapped_gate": to_dict(gate.wrapped_gate),
    }


@to_dict.register
def _exponential_gate_to_dict(gate: _gates.Exponential):
    return {
        "name": gate.name,
        "wrapped_gate": to_dict(gate.wrapped_gate),
    }


@to_dict.register
def _power_gate_to_dict(gate: _gates.Power):
    return {
        "name": gate.name,
        "wrapped_gate": to_dict(gate.wrapped_gate),
        "exponent": gate.exponent,
    }


# ---------- deserialization ----------


def circuit_from_dict(dict_):
    defs = [
        custom_gate_def_from_dict(def_dict)
        for def_dict in dict_.get("custom_gate_definitions", [])
    ]
    return _circuit.Circuit(
        operations=[
            _gate_operation_from_dict(op_dict, defs)
            for op_dict in dict_.get("operations", [])
        ],
        n_qubits=dict_["n_qubits"],
    )


def _gate_operation_from_dict(dict_, custom_gate_defs):
    return _gates.GateOperation(
        gate=_gate_from_dict(dict_["gate"], custom_gate_defs),
        qubit_indices=tuple(dict_["qubit_indices"]),
    )


def _gate_from_dict(dict_, custom_gate_defs):
    """Generic gate deserializer.

    Pass it a JSON dictionary and it'll try its best to return you a proper gate
    instance, regardless of the given gate type."""
    try:
        return _builtin_gate_from_dict(dict_)
    except KeyError:
        pass

    try:
        return _special_gate_from_dict(dict_, custom_gate_defs)
    except KeyError:
        pass

    return _custom_gate_instance_from_dict(dict_, custom_gate_defs)


def _builtin_gate_from_dict(dict_) -> _builtin_gates.GateRef:
    gate_ref = builtin_gate_by_name(dict_["name"])
    if gate_ref is None:
        raise KeyError()

    if _gates.gate_is_parametric(gate_ref, dict_.get("params")):
        return gate_ref(
            *[
                deserialize_expr(param, dict_.get("free_symbols", []))
                for param in dict_.get("params", [])
            ]
        )
    else:
        return gate_ref


def _special_gate_from_dict(dict_, custom_gate_defs) -> _gates.Gate:
    if dict_["name"] == _gates.CONTROLLED_GATE_NAME:
        wrapped_gate = _gate_from_dict(dict_["wrapped_gate"], custom_gate_defs)
        return _gates.ControlledGate(wrapped_gate, dict_["num_control_qubits"])

    elif dict_["name"].endswith(_gates.DAGGER_GATE_NAME):
        wrapped_gate = _gate_from_dict(dict_["wrapped_gate"], custom_gate_defs)
        return _gates.Dagger(wrapped_gate)

    elif dict_["name"] == _gates.EXPONENTIAL_GATE_NAME:
        wrapped_gate = _gate_from_dict(dict_["wrapped_gate"], custom_gate_defs)
        return _gates.Exponential(wrapped_gate)

    elif _gates.POWER_GATE_SYMBOL in dict_["name"]:
        wrapped_gate = _gate_from_dict(dict_["wrapped_gate"], custom_gate_defs)
        return _gates.Power(wrapped_gate, dict_["exponent"])

    else:
        raise KeyError()


def custom_gate_def_from_dict(dict_) -> _gates.CustomGateDefinition:
    symbols = [sympy.Symbol(term) for term in dict_.get("params_ordering", [])]
    return _gates.CustomGateDefinition(
        gate_name=dict_["gate_name"],
        matrix=_matrix_from_json(dict_["matrix"], dict_.get("params_ordering", [])),
        params_ordering=tuple(symbols),
    )


def _custom_gate_instance_from_dict(dict_, custom_gate_defs) -> _gates.Gate:
    gate_def = next(
        (
            gate_def
            for gate_def in custom_gate_defs
            if gate_def.gate_name == dict_["name"]
        ),
        None,
    )
    if gate_def is None:
        raise ValueError(
            f"Custom gate definition for {dict_['name']} missing from serialized dict"
        )

    symbol_names = map(serialize_expr, gate_def.params_ordering)
    return gate_def(
        *[deserialize_expr(param, symbol_names) for param in dict_.get("params", [])]
    )


def circuitset_from_dict(dict_) -> List[_circuit.Circuit]:
    return _map_eager(circuit_from_dict, dict_["circuits"])


def load_circuit(load_src: LoadSource):
    with ensure_open(load_src) as f:
        return circuit_from_dict(json.load(f))


def load_circuitset(load_src: LoadSource):
    with ensure_open(load_src) as f:
        return circuitset_from_dict(json.load(f))


def save_circuit(circuit: _circuit.Circuit, dump_target: DumpTarget):
    with ensure_open(dump_target, "w") as f:
        json.dump(to_dict(circuit), f)


def save_circuitset(circuitset: List[_circuit.Circuit], dump_target: DumpTarget):
    with ensure_open(dump_target, "w") as f:
        json.dump(to_dict(circuitset), f)



################################################################################
# © Copyright 2021-2022 Zapata Computing Inc.
################################################################################
from typing import Optional

from orquestra.quantum.api.wavefunction_simulator import BaseWavefunctionSimulator
from orquestra.quantum.circuits import Circuit, Operation

from ..typing import StateVector


class SymbolicSimulator(BaseWavefunctionSimulator):
    """A simulator computing wavefunction by consecutive gate matrix multiplication.

    Args:
        seed: the seed of the sampler
    """

    def __init__(
        self,
        *,
        seed: Optional[int] = None,
    ):
        super().__init__(seed=seed)

    def _get_wavefunction_from_native_circuit(
        self, circuit: Circuit, initial_state: StateVector
    ) -> StateVector:
        state = initial_state

        for operation in circuit.operations:
            state = operation.apply(state)

        return state

    def is_natively_supported(self, operation: Operation) -> bool:
        return True



################################################################################
# © Copyright 2022 Zapata Computing Inc.
################################################################################
from itertools import product
from typing import List
from unittest.mock import Mock

import pytest

from orquestra.quantum.circuits import Circuit, X, Y, Z
from orquestra.quantum.operators._pauli_operators import PauliSum, PauliTerm


@pytest.fixture
def pauli_term():
    correct_list = [("X", 0), ("Y", 1), ("Z", 12)]
    return PauliTerm.from_iterable(correct_list, coefficient=2.0)


@pytest.fixture
def pauli_sum():
    return 0.5 * PauliTerm("X0") + 0.5j * PauliTerm("Y0")


class TestPauliTermInitialization:
    @pytest.mark.parametrize(
        "operator_dict, coefficient",
        [({1: "X"}, 1.5), ({0: "X", 1: "X"}, 0.5), ({0: "Y", 10: "Z"}, 0.5 + 0.5j)],
    )
    def test_term_can_be_initialized_with_dictionary(self, operator_dict, coefficient):
        term = PauliTerm(operator_dict, coefficient)
        assert term.operations == frozenset(operator_dict.items())
        assert term.coefficient == coefficient
        assert term.qubits == frozenset(operator_dict)

    @pytest.mark.parametrize(
        "pauli_str, coefficient, qubit_index, operator",
        [
            ("X0", 1.0, 0, "X"),
            ("Z0", -1j, 0, "Z"),
            ("z1", -2.0, 1, "Z"),
            ("y2", 1.5, 2, "Y"),
            ("X123", 4 + 0.5j, 123, "X"),
        ],
    )
    def test_term_can_be_initialized_using_single_operator_and_coefficient(
        self, pauli_str, coefficient, qubit_index, operator
    ):
        term = PauliTerm(pauli_str, coefficient)
        assert term.operations == frozenset([(qubit_index, operator)])
        assert term.coefficient == coefficient
        assert term.qubits == {qubit_index}

    @pytest.mark.parametrize(
        "pauli_str, coefficient, qubit_indices, operators",
        [
            ("X0 * Z1", -1.0, (0, 1), ("X", "Z")),
            ("X3 * Y12", -0.5, (3, 12), ("X", "Y")),
        ],
    )
    def test_term_can_be_initialized_using_multiple_operators_and_coefficient(
        self, pauli_str, coefficient, qubit_indices, operators
    ):
        term = PauliTerm(pauli_str, coefficient)
        assert term.operations == frozenset(zip(qubit_indices, operators))
        assert term.coefficient == coefficient
        assert term.qubits == frozenset(qubit_indices)

    @pytest.mark.parametrize(
        "pauli_str, coefficient, qubit_indices, operators",
        [
            ("-1.0 * X0 * Z1", -1.0, (0, 1), ("X", "Z")),
            ("-0.5 * X3 * Y12", -0.5, (3, 12), ("X", "Y")),
            ("(1-0.5j) * X3 * Y12", 1 - 0.5j, (3, 12), ("X", "Y")),
            ("(2 + 3j) * y0 * z2", 2 + 3j, (0, 2), ("Y", "Z")),
        ],
    )
    def test_term_can_be_initialized_by_passing_operators_and_coef_in_string(
        self, pauli_str, coefficient, qubit_indices, operators
    ):
        term = PauliTerm(pauli_str)
        assert term.operations == frozenset(zip(qubit_indices, operators))
        assert term.coefficient == coefficient
        assert term.qubits == frozenset(qubit_indices)

    @pytest.mark.parametrize(
        "pauli_str, expected_ops",
        [
            ("I0", {}),
            ("I12", {}),
            ("X0 * I1 * I34", {(0, "X")}),
            ("X0 * Z1 * I34", {(0, "X"), (1, "Z")}),
        ],
    )
    def test_identity_operators_are_not_stored_in_iterm(self, pauli_str, expected_ops):
        assert PauliTerm(pauli_str).operations == frozenset(expected_ops)

    @pytest.mark.parametrize(
        "pauli_str", ["1 X", "X - 1254", "A0", "5.0 + Z1", "0.3+0.5j * X0", "X0 Z1"]
    )
    def test_term_cannot_be_constructed_from_badly_formatted_string(self, pauli_str):
        with pytest.raises(ValueError) as e:
            PauliTerm(pauli_str)

        assert "Badly formatted" in str(e.value)

    @pytest.mark.parametrize("operator_dict", [{0: "XX"}, {0: "X", 1: "A"}])
    def test_term_cannot_be_constructed_from_dictionary_containing_incorrect_operators(
        self, operator_dict
    ):
        with pytest.raises(ValueError):
            PauliTerm(operator_dict)

    @pytest.mark.parametrize("pauli_str", ["X0 * Y0", "X0 * Y1 * Z1"])
    def test_term_cannot_be_constructed_if_duplicate_qubits_are_present_in_pauli_str(
        self, pauli_str
    ):
        with pytest.raises(ValueError):
            PauliTerm(pauli_str)

    @pytest.mark.parametrize(
        "pauli_str, coefficient", [("0.5 * X0", 0.5), ("2.0 * Y12", 3.0)]
    )
    def test_term_cannot_be_constructed_if_coefficient_is_passed_via_both_ways(
        self, pauli_str, coefficient
    ):
        with pytest.raises(ValueError) as e:
            PauliTerm(pauli_str, coefficient)

        assert "Coefficient can be provided either in an argument " in str(e.value)

    @pytest.mark.parametrize(
        "term",
        [
            PauliTerm("X0", -5.0),
            PauliTerm("-5.0 * Z1"),
            PauliTerm("2.0 * X0 * Y2 * Z3"),
            PauliTerm("X0 * Y2 * Z3", 2.0),
        ],
    )
    def test_creating_term_from_its_str_representation_gives_the_same_term(self, term):
        assert PauliTerm(str(term)) == term


class TestConstructingPauliTermFromIterable:
    @pytest.mark.parametrize(
        "ops, coefficient",
        [
            ([("X", 0), ("Y", 1), ("Z", 12)], -0.5),
            ({("X", 0), ("Y", 1), ("Z", 12)}, -0.5),
        ],
    )
    def test_term_constructed_from_iterable_has_expected_operators_and_coefficient(
        self, ops, coefficient
    ):
        term = PauliTerm.from_iterable(ops, coefficient)
        # The reverse (pair[::-1]) is from the fact that from_iterable accepts input in
        # the reverse direction then the operations produces it.
        assert term.operations == frozenset([pair[::-1] for pair in ops])
        assert term.coefficient == coefficient

    def test_identity_term_is_constructed_from_empty_iterable(self):
        term = PauliTerm.from_iterable([])
        assert term.operations == frozenset()
        assert term.coefficient == 1.0
        assert term == PauliTerm.identity()

        term_with_coeff = PauliTerm.from_iterable([], 2.0 + 3.0j)
        assert term_with_coeff.operations == frozenset()
        assert term_with_coeff.coefficient == 2.0 + 3.0j
        assert term_with_coeff == PauliTerm.identity() * (2.0 + 3.0j)

    def test_term_cannot_be_constructed_from_iterable_of_incorrectly_shaped_tuples(
        self,
    ):
        with pytest.raises(ValueError) as e:
            PauliTerm.from_iterable([("X0")])

        assert "list can only contain" in str(e.value)

    def test_term_cannot_be_initialized_if_any_index_in_list_is_incorrect(self):
        with pytest.raises(ValueError) as e:
            PauliTerm.from_iterable([("X", -1)])

        assert "Invalid qubit index" in str(e.value)

    def test_term_cannot_be_constructed_if_qubit_indices_are_duplicate(self):
        with pytest.raises(ValueError) as e:
            PauliTerm.from_iterable([("X", 0), ("Y", 0)])
        assert "Duplicate" in str(e.value)


class TestPauliTermIdentity:
    def test_identity_stores_coefficient_of_one_and_no_operators(self):
        identity_term = PauliTerm.identity()

        assert len(identity_term) == 0
        assert identity_term.coefficient == 1.0

    @pytest.mark.parametrize(
        "term",
        [PauliTerm("X1"), PauliTerm("X0 * X1", -0.5), PauliTerm("(2.5+1j) * X0 * Z12")],
    )
    def test_identity_is_neutral_element_for_pauli_terms_multiplication(self, term):
        assert PauliTerm.identity() * term == term
        assert term * PauliTerm.identity() == term


class TestCopyingPauliTerm:
    def test_copying_term_returns_new_object(self):
        original = PauliTerm("X0", 2.0)
        copy = original.copy(new_coefficient=3.0)

        assert copy is not original
        assert copy._ops is not original._ops

    def test_copy_of_term_has_the_same_coef_if_new_coef_is_not_provided(self):
        assert PauliTerm("(2.0+3j) * Y12").copy().coefficient == 2.0 + 3j

    def test_copying_terms_with_new_coefficient_does_not_modify_original_term(self):
        original = PauliTerm("Y0 * Z10", -1.0)
        copy = original.copy(new_coefficient=2.5)

        assert original.coefficient == -1.0
        assert copy.coefficient == 2.5


class TestPauliOperatorProperties:
    @pytest.mark.parametrize(
        "term",
        [
            PauliTerm("Z0", -2),
            PauliTerm("-0.5 * Z1 * Z3"),
            PauliTerm.from_iterable([("Z", 3), ("Z", 4)], 5.0),
        ],
    )
    def test_term_is_ising_if_it_contains_only_z_operators(self, term):
        assert term.is_ising

    @pytest.mark.parametrize(
        "term", [PauliTerm("X0"), PauliTerm("Z0 * X1"), PauliTerm("X1 * Y2 * Z3")]
    )
    def test_term_is_not_ising_if_any_of_its_nontrivial_operators_are_not_z(self, term):
        assert not term.is_ising

    @pytest.mark.parametrize(
        "term", [PauliTerm("I0", 2.0), PauliTerm({}), PauliTerm({0: "I", 1: "I"})]
    )
    def test_term_is_constant_if_it_comprises_only_trivial_operators(self, term):
        assert term.is_constant

    @pytest.mark.parametrize(
        "term",
        [
            PauliTerm("X0"),
            PauliTerm("2.5 * X0 * Z2 * Y10"),
            PauliTerm("0.5 * I0 * I1 * Z3"),
        ],
    )
    def test_term_is_not_constant_if_it_comprises_nontrivial_operators(self, term):
        assert not term.is_constant

    @pytest.mark.parametrize(
        "operator,n_qubits",
        [
            (PauliTerm("X0"), 1),
            (PauliSum("2"), 0),
            (PauliTerm("Z0*Z3"), 4),
            (PauliTerm("X0*X1"), 2),
            (PauliSum("2*Z2 + X5"), 6),
        ],
    )
    def test_n_qubits(self, operator, n_qubits):
        assert operator.n_qubits == n_qubits


class TestPauliTermToCircuitConversion:
    def test_circuit_constructed_from_pauli_term_is_cached(self):
        term = PauliTerm("X0 * Z1 * Y3", -1.0)
        assert not hasattr(term, "_circuit")
        circuit = term.circuit
        assert hasattr(term, "_circuit")
        assert term.circuit is circuit

    def test_circuit_obtained_from_term_has_correct_gates(self):
        term = PauliTerm("-3.5 * X0 * Z1 * Y3")
        assert term.circuit == Circuit([X(0), Z(1), Y(3)])


class TestPauliTermIndexingAndIteration:
    @pytest.mark.parametrize(
        "term", [PauliTerm("X0 * I2 * Y3 * Z4"), PauliTerm("X0 * Y3 * Z4")]
    )
    def test_accessing_undefined_or_trivial_indices_warns_and_returns_identity(
        self, term
    ):
        with pytest.warns(UserWarning):
            returned_op = term[2]

        assert returned_op == "I"

    def test_iterative_over_terms_yields_correct_pairs_of_operator_and_index(self):
        term = PauliTerm("X0 * Y1 * Z12")
        expected_items = [("X", 0), ("Y", 1), ("Z", 12)]

        assert list(term) == expected_items


class TestPauliTermAlgebra:
    def test_pauliterm_equality(self, pauli_term):
        # Check copying creates an equal term
        copied_term = pauli_term.copy()

        assert copied_term == pauli_term
        assert pauli_term == PauliSum([copied_term])

        # Check __eq__ of PauliSum is called
        old_paulisum_eq = PauliSum.__eq__
        PauliSum.__eq__ = Mock()
        pauli_term == PauliSum([copied_term])
        PauliSum.__eq__.assert_called()
        PauliSum.__eq__.assert_called_with(pauli_term)
        PauliSum.__eq__ = old_paulisum_eq

        # Check removing a term makes the terms unequal
        copied_term *= PauliTerm("X0")

        assert copied_term != pauli_term

        # Check constant term is comparable with numbers
        assert PauliTerm.identity() == 1.0

        # Check error is raised for invalid type
        with pytest.raises(TypeError):
            pauli_term == "X0*Y1*Z12"

    def test_pauliterm_equality_zero(self, pauli_term: PauliTerm):
        # Test that various representations of zero are equal
        zero_paulisum = PauliSum()
        zero_identity_term = PauliTerm("I0", 0)
        mul_0 = pauli_term * 0
        copy_0 = pauli_term.copy(0)
        other_0 = PauliTerm("X5*Z11", 0)
        paulisum_with_zero_term = PauliSum([zero_identity_term]).simplify()
        subtracted = pauli_term - pauli_term

        assert (
            zero_paulisum
            == zero_identity_term
            == mul_0
            == copy_0
            == other_0
            == paulisum_with_zero_term
            == subtracted
        )

    def test_pauliterm_add(self, pauli_term):
        # Test adding to PauliSum calls function of PauliSum
        old_paulisum_add = PauliSum.__add__
        PauliSum.__add__ = Mock()
        pauli_term + PauliSum.identity()
        PauliSum.__add__.assert_called_with(pauli_term)
        PauliSum.__add__ = old_paulisum_add

        # Test adding to term creates PauliSum with both terms
        summation = pauli_term + PauliTerm.identity()
        assert len(summation) == 2
        assert pauli_term in summation
        assert PauliTerm.identity() in summation

        # Test adding to a constant
        summation = pauli_term + 3.0
        assert PauliTerm("I0", 3.0) in summation

    def test_pauliterm_radd(self, pauli_term):
        summation = 3.0 + pauli_term
        assert PauliTerm("I0", 3.0) in summation

        old_pauliterm_add = PauliTerm.__add__
        PauliTerm.__add__ = Mock()
        3.0 + pauli_term
        PauliTerm.__add__.assert_called_once_with(3.0)
        PauliTerm.__add__ = old_pauliterm_add

    def test_pauliterm_sub(self, pauli_term):
        assert pauli_term - PauliTerm("X3") == pauli_term + PauliTerm("X3", -1)

        assert pauli_term - pauli_term == PauliSum()

    def test_pauliterm_rsub(self):
        assert 3.0 - PauliTerm("X3") == 3.0 + PauliTerm("X3", -1)

    def test_pauliterm_mul(self, pauli_term):
        # Test constant just creates similar term with multiplied coeffs
        pauli_term.coefficiet = 2.0
        constant_mul = pauli_term * 2.0j
        assert constant_mul._ops == pauli_term._ops
        assert constant_mul.coefficient == 4.0j
        assert constant_mul is not pauli_term

        # Test mul with PauliSum delegates to mul of PauliSum
        old_paulisum_mul = PauliSum.__mul__
        PauliSum.__mul__ = Mock()
        pauli_term * PauliSum.identity()
        PauliSum.__mul__.assert_called_once_with(PauliSum.identity())
        PauliSum.__mul__ = old_paulisum_mul

        # Test different scenarios for PauliTerm
        for combination in product(("X", "Y", "Z", "I"), repeat=2):
            left, right = combination[0], combination[1]

            result = PauliTerm(f"{left}0") * PauliTerm(f"{right}0")

            if left == "I" or right == "I":
                other_op = left if right == "I" else right
                assert result == PauliTerm(f"{other_op}0")
                continue

            if left == right:
                assert result == PauliTerm.identity()
                continue

            if "X" in combination and "Y" in combination:
                assert result._ops[0] == "Z"

                if left == "X":
                    assert result.coefficient == 1.0j
                else:
                    assert result.coefficient == -1.0j

            if "Y" in combination and "Z" in combination:
                assert result._ops[0] == "X"

                if left == "Y":
                    assert result.coefficient == 1.0j
                else:
                    assert result.coefficient == -1.0j

            if "X" in combination and "Z" in combination:
                assert result._ops[0] == "Y"

                if left == "X":
                    assert result.coefficient == -1.0j
                else:
                    assert result.coefficient == 1.0j

    def test_pauliterm_rmul(self, pauli_term):
        multiplication = 3.0 * pauli_term
        assert multiplication.coefficient == 6.0

        old_pauliterm_mul = PauliTerm.__mul__  # Need to recover later
        PauliTerm.__mul__ = Mock(return_value=PauliTerm.identity())
        3.0 * pauli_term
        PauliTerm.__mul__.assert_called_with(3.0)
        PauliTerm.__mul__ = old_pauliterm_mul

    def test_pauliterm_pow(self, pauli_term):
        with pytest.raises(ValueError):
            pauli_term**-10
        with pytest.raises(ValueError):
            pauli_term**0.25

        assert pauli_term**0 == PauliTerm.identity()
        assert pauli_term**1 == pauli_term

    def test_pauliterm_repr(self, pauli_term: PauliTerm):
        # Check identity is printed correctly (integer coefficients)
        assert str(PauliTerm.identity()) == "1.0*I"

        # Check identity is printed correctly (decimal coefficients)
        assert (
            str(PauliTerm.identity().copy(new_coefficient=1.5 + 2.3j)) == "(1.5+2.3j)*I"
        )

        # Check Pauli Term is printed correctly (integer coefficients)
        assert str(pauli_term) == "2.0*X0*Y1*Z12"

        # Check Pauli Term is printed correctly (decimal coefficients)
        assert (
            str(pauli_term.copy(new_coefficient=3.3 + 5.2j)) == "(3.3+5.2j)*X0*Y1*Z12"
        )

    def test_pauliterm_multiply_by_zero(self, pauli_term: PauliTerm):
        zero_op = pauli_term * 0
        assert isinstance(zero_op, PauliTerm)
        assert zero_op.coefficient == 0


class TestPauliSumOperations:
    @pytest.fixture
    def pauli_sum(self):
        return (
            0.5 * PauliTerm("X0")
            + 0.5j * PauliTerm("Y0")
            + PauliTerm("Z1") * PauliTerm("Z2")
        )

    def test_paulisum_native_constructor(self):
        # Empty parameters list returns empty sum
        assert PauliSum().terms == []

        # Incorrect list contents
        with pytest.raises(ValueError):
            PauliSum([PauliTerm.identity(), 1])

    def test_paulisum_cannot_be_constructed_from_term_not_wrapped_in_list(self):
        with pytest.raises(ValueError):
            PauliSum(PauliTerm.identity())

    def test_paulisum_fromstr(self, pauli_sum):
        assert PauliSum(str(pauli_sum)) == pauli_sum

    def test_paulisum_qubits(self, pauli_sum):
        assert pauli_sum.qubits == {0, 1, 2}

    def test_paulisum_isising(self, pauli_sum):
        assert not pauli_sum.is_ising

        ising_sum = PauliSum(
            [PauliTerm("Z1") * PauliTerm("Z2"), PauliTerm("Z12"), PauliTerm("Z34")]
        )

        assert ising_sum.is_ising

    def test_paulisum_circuits(self, pauli_sum):
        assert not hasattr(pauli_sum, "_circuits")
        circuits: List[Circuit] = pauli_sum.circuits
        assert hasattr(pauli_sum, "_circuits")

        for circuit, term in zip(circuits, pauli_sum):
            for op in circuit.operations:
                assert op.gate.name == term[op.qubit_indices[0]]

    def test_paulisum_identity(self):
        sum_id = PauliSum.identity()

        assert len(sum_id) == 1
        assert sum_id[0] == PauliTerm.identity()

    def test_paulisum_eq(self, pauli_sum):
        assert PauliSum.identity() == 1.0
        assert PauliSum.identity() == PauliTerm.identity()

        assert pauli_sum != PauliSum.identity()
        assert pauli_sum != PauliSum(pauli_sum[:-1])

        assert pauli_sum == PauliSum(pauli_sum[::-1])

        with pytest.raises(TypeError):
            pauli_sum == "(0.5)*X0 + (0.5j)*Y0 + (1.0)*Z1*Z2"

    def test_paulisum_add(self, pauli_sum):
        # Check simplify is called during sum
        old_paulisum_simplify = PauliSum.simplify
        PauliSum.simplify = Mock()
        pauli_sum + pauli_sum
        PauliSum.simplify.assert_called_once()
        PauliSum.simplify = old_paulisum_simplify

        # Check simplify works correctly
        summation = pauli_sum + pauli_sum
        assert len(summation) == len(pauli_sum)
        for term in pauli_sum:
            assert str(term.copy(new_coefficient=term.coefficient * 2)) in str(
                summation
            )

        # Check summation with constant
        constant_sum = PauliSum.identity() + 3.0
        assert len(constant_sum) == 1
        assert constant_sum[0].coefficient == 4

        # Check summation with pauliterm
        constant_sum = PauliSum.identity() + PauliTerm.identity()
        assert len(constant_sum) == 1
        assert constant_sum[0].coefficient == 2

    def test_paulisum_radd(self, pauli_sum):
        old_paulisum_radd = PauliSum.__radd__
        old_paulisum_add = PauliSum.__add__

        PauliSum.__radd__ = Mock(side_effect=PauliSum.__radd__)
        PauliSum.__add__ = Mock()

        # Actual call
        PauliSum.__radd__(pauli_sum, other=5)

        PauliSum.__radd__.assert_called_once()
        PauliSum.__add__.assert_called_once_with(5)

        PauliSum.__radd__ = old_paulisum_radd
        PauliSum.__add__ = old_paulisum_add

    def test_paulisum_sub(self, pauli_sum):
        assert pauli_sum - pauli_sum == PauliSum()

    def test_paulisum_rsub(self, pauli_sum):
        subtraction = 0 - pauli_sum

        for neg_term, term in zip(subtraction, pauli_sum):
            assert neg_term.coefficient == -1 * term.coefficient

    def test_paulisum_mul(self, pauli_sum):
        multiplication = pauli_sum * PauliTerm("X0")

        for mul_term, term in zip(multiplication, pauli_sum):
            assert mul_term == term * PauliTerm("X0")

        assert pauli_sum * 5 == 5 * pauli_sum

    def test_paulisum_rmul(self, pauli_sum):
        mul_by_constant = 5.0j * pauli_sum

        assert all(
            [
                mul_term.coefficient == term.coefficient * 5j
                for mul_term, term in zip(mul_by_constant, pauli_sum)
            ]
        )

    def test_paulisum_pow(self, pauli_sum):
        with pytest.raises(ValueError):
            pauli_sum**-1
        with pytest.raises(ValueError):
            pauli_sum**0.25

        assert pauli_sum**0 == PauliTerm.identity()
        assert pauli_sum**1 == pauli_sum

    def test_paulisum_repr(self, pauli_sum):
        repr_str = str(pauli_sum)

        assert all([str(term) in repr_str for term in pauli_sum.terms])

    def test_paulisum_multiply_by_zero(self, pauli_sum: PauliSum):
        zero_op = pauli_sum * 0
        assert zero_op == PauliSum()

    def test_paulisum_constant_term_zero(self, pauli_sum):
        assert pauli_sum.constant_term == 0

    @pytest.mark.parametrize("constant_term", [10, -10, 0.5, -0.5])
    def test_paulisum_constant(self, pauli_sum, constant_term):
        pauli_sum = pauli_sum + constant_term
        assert pauli_sum.constant_term == constant_term

    @pytest.mark.parametrize("constant_term", [10, -10, 0.5, -0.5])
    def test_paulisum_identity_operator(self, pauli_sum, constant_term):
        pauli_sum = pauli_sum + constant_term * PauliTerm("I0")
        assert pauli_sum.constant_term == constant_term
